# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for gwsumm.html.bootstrap
"""

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

import pytest
from datetime import datetime

from gwdetchar.utils import parse_html

from .. import bootstrap

# global variables
DATE = datetime.strptime('20140410', '%Y%m%d')
BACKWARD = '<a class="nav-link step-back" title="Step backward">&laquo;</a>'
CAL = ('<a id="calendar" class="nav-link dropdown-toggle" title="Show/hide '
       'calendar" data-date="10-04-2014" data-date-format="dd-mm-yyyy" '
       'data-viewmode="{}">{}</a>')
FORWARD = '<a class="nav-link step-forward" title="Step forward">&raquo;</a>'


# test utilities

def test_banner():
    banner = bootstrap.banner('Test', subtitle='Subtest')
    assert parse_html(str(banner)) == parse_html(
        '<div class="banner">\n<h1>Test</h1>\n<p>Subtest</p>\n</div>')
    # test with classes
    banner_wclass = bootstrap.banner(
        'Test', subtitle='Subtest', titleclass='test', subtitleclass='subtest')
    assert parse_html(str(banner_wclass)) == parse_html(
        '<div class="banner">\n<h1 class=\"test\">Test</h1>\n'
        '<p class=\"subtest\">Subtest</p>\n</div>')


@pytest.mark.parametrize('mode, datefmt', [
    ('day', 'April 10 2014'),
    ('week', 'Week of April 10 2014'),
    ('month', 'April 2014'),
    ('year', '2014'),
])
def test_calendar(mode, datefmt):
    backward, cal, forward = bootstrap.calendar(DATE, mode=mode)
    assert parse_html(str(backward)) == parse_html(str(BACKWARD))
    assert parse_html(str(cal)) == parse_html(
        CAL.format('%ss' % mode, datefmt))
    assert parse_html(str(forward)) == parse_html(str(FORWARD))


def test_calendar_no_mode():
    # test with no Mode
    with pytest.raises(ValueError) as exc:
        bootstrap.calendar(DATE)
    assert str(exc.value).startswith('Cannot generate calendar for Mode')


def test_wrap_content():
    content = bootstrap.wrap_content('test')
    assert parse_html(str(content)) == parse_html(
        '<div class="container-fluid" id="main">\ntest\n</div>')


def test_state_switcher():
    switcher = bootstrap.state_switcher([('Test', '#test')])
    assert parse_html(str(switcher)) == parse_html(
        '<ul class="nav navbar-nav">\n<li class="nav-item dropdown">\n'
        '<a class="nav-link dropdown-toggle" href="#" id="states" role='
        '"button" title="Show/hide state menu" data-toggle="dropdown">Test</a>'
        '\n<div class="dropdown-menu dropdown-menu-right state-switch shadow" '
        'id="statemenu">\n<h6 class="dropdown-header">Select below to view '
        'this page in another state (different time segments).</h6>\n<div '
        'class="dropdown-divider"></div>\n<a class="dropdown-item state" '
        'title="Test" id="state_test" onclick="jQuery(this).load_state'
        '(&quot;#test&quot;);">Test</a>\n</div>\n</li>\n</ul>')


def test_base_map_dropdown():
    menu = bootstrap.base_map_dropdown('test', id_='id')
    assert parse_html(str(menu)) == parse_html(
        '<div class="navbar-brand border border-white '
        'rounded" id="id">test</div>')
    # test with bases
    menu_wbases = bootstrap.base_map_dropdown('test', bases={'key': 'value'})
    assert parse_html(str(menu_wbases)) == parse_html(
        '<div class="dropdown base-map">\n<a href="#" class="navbar-brand '
        'nav-link border border-white rounded dropdown-toggle" role="button" '
        'data-toggle="dropdown">test</a>\n<div class="dropdown-menu '
        'dropdown-1-col shadow">\n<a title="key" class="dropdown-item" '
        'data-new-base="value">key</a>\n</div>\n</div>')



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Read and store transient event triggers
"""

import warnings
from urllib.parse import urlparse

from astropy.table import vstack as vstack_tables

from lal.utils import CacheEntry

from ligo.lw import lsctables

from glue.lal import Cache

from gwpy.io.cache import cache_segments
from gwpy.table import (EventTable, filters as table_filters)
from gwpy.table.filter import parse_column_filters
from gwpy.table.io.pycbc import filter_empty_files as filter_pycbc_live_files
from gwpy.segments import (DataQualityFlag, SegmentList)

import gwtrigfind

from . import globalv
from .utils import (re_cchar, vprint, safe_eval)
from .config import GWSummConfigParser
from .channels import get_channel

# build list of default keyword arguments for reading ETGs
ETG_READ_KW = {
    'cwb': {
        'format': 'root',
        'treename': 'waveburst',
    },
    'daily_ihope': {
        'format': 'ligolw',
        'tablename': 'sngl_inspiral',
        'use_numpy_dtypes': True,
    },
    'daily_ahope': {
        'format': 'ligolw',
        'tablename': 'sngl_inspiral',
        'use_numpy_dtypes': True,
    },
    'dmt_omega': {
        'format': 'ligolw',
        'tablename': 'sngl_burst',
        'use_numpy_dtypes': True,
    },
    'dmt_wsearch': {
        'format': 'ligolw',
        'tablename': 'sngl_burst',
        'use_numpy_dtypes': True,
    },
    'kleinewelle': {
        'format': 'ligolw',
        'tablename': 'sngl_burst',
        'use_numpy_dtypes': True,
    },
    'kw': {
        'format': 'ligolw',
        'tablename': 'sngl_burst',
        'use_numpy_dtypes': True,
    },
    'omega': {
        'format': 'ascii',
    },
    'omegadq': {
        'format': 'ascii',
    },
    'omicron': {
        'format': 'ligolw',
        'tablename': 'sngl_burst',
        'use_numpy_dtypes': True,
    },
    'pycbc_live': {
        'format': 'hdf5.pycbc_live',
        'timecolumn': 'end_time',
        'extended_metadata': False,
    },
}

# set default for all LIGO_LW
for name in lsctables.TableByName:
    ETG_READ_KW[name] = {
        'format': 'ligolw',
        'tablename': name,
        'use_numpy_dtypes': True,
    }


def get_etg_table(etg):
    """Find which table should be used for the given etg

    Parameters
    ----------
    etg : `str`
        name of Event Trigger Generator for which to query

    Returns
    -------
    table : `type`, subclass of `~ligo.lw.table.Table`
        LIGO_LW table registered to the given ETG

    Raises
    ------
    KeyError
        if the ETG is not registered
    """
    try:
        kw_ = get_etg_read_kwargs(etg)
        form = kw_['format']
        tablename = kw_['tablename']
    except KeyError as e:
        e.args = ('No LIGO_LW table registered to etg %r' % etg,)
        raise
    if form == 'ligolw':
        return lsctables.TableByName[tablename]
    raise KeyError("No LIGO_LW table registered to etg %r" % etg)


def get_triggers(channel, etg, segments, config=GWSummConfigParser(),
                 cache=None, columns=None, format=None, query=True,
                 nproc=1, ligolwtable=None, filter=None,
                 timecolumn=None, verbose=False, return_=True):
    """Read a table of transient event triggers for a given channel.
    """
    key = '%s,%s' % (str(channel), etg.lower())
    # convert input segments to a segmentlist (for convenience)
    if isinstance(segments, DataQualityFlag):
        segments = segments.active
    segments = SegmentList(segments)

    # get read keywords for this etg
    read_kw = get_etg_read_kwargs(etg, config=config, exclude=[])
    read_kw['verbose'] = verbose

    # extract columns (using function keyword if given)
    if columns:
        read_kw['columns'] = columns
    columns = read_kw.pop('columns', None)

    # override with user options
    if format:
        read_kw['format'] = format
    elif not read_kw.get('format', None):
        read_kw['format'] = etg.lower()
    if timecolumn:
        read_kw['timecolumn'] = timecolumn
    elif columns is not None and 'time' in columns:
        read_kw['timecolumn'] = 'time'

    # replace columns keyword
    if read_kw['format'].startswith('ascii.'):
        read_kw['include_names'] = columns
    else:
        read_kw['columns'] = columns

    # parse filters
    if filter:
        read_kw['selection'].extend(parse_column_filters(filter))

    # read segments from global memory
    try:
        havesegs = globalv.TRIGGERS[key].meta['segments']
    except KeyError:
        new = segments
    else:
        new = segments - havesegs

    # read new triggers
    if query and abs(new) != 0:
        ntrigs = 0
        vprint("    Grabbing %s triggers for %s" % (etg, str(channel)))

        # -- setup ----------

        # get find/read kwargs
        trigfindkwargs = dict(
             (k[9:], read_kw.pop(k)) for k in list(read_kw) if
             k.startswith('trigfind-'))
        trigfindetg = trigfindkwargs.pop('etg', etg)

        # customise kwargs for this ETG
        if etg.lower().replace('-', '_') in ['pycbc_live']:
            read_kw['ifo'] = get_channel(channel).ifo
        if etg.lower() in ['kw', 'kleinewelle']:
            read_kw['selection'].append('channel == "%s"' % channel)
        if etg.lower() in ['cwb'] and 'root' not in read_kw['format']:
            read_kw.pop('treename')

        # filter on segments
        if 'timecolumn' in read_kw:
            read_kw['selection'].append((
                read_kw['timecolumn'], table_filters.in_segmentlist, new))

        # -- read -----------

        # if single file
        if cache is not None and len(cache) == 1:
            trigs = read_cache(cache, new, etg, nproc=nproc, **read_kw)
            if trigs is not None:
                add_triggers(trigs, key)
                ntrigs += len(trigs)
        # otherwise, loop over segments
        else:
            for segment in new:
                # find trigger files
                if cache is None and not etg.lower() == 'hacr':
                    try:
                        segcache = gwtrigfind.find_trigger_files(
                            str(channel), trigfindetg, segment[0], segment[1],
                            **trigfindkwargs)
                    except ValueError as e:
                        warnings.warn("Caught %s: %s"
                                      % (type(e).__name__, str(e)))
                        continue
                elif cache is not None:
                    segcache = cache
                # read table
                if etg.lower() == 'hacr':
                    from gwpy.table.io.hacr import get_hacr_triggers
                    trigs = get_hacr_triggers(channel, segment[0], segment[1],
                                              columns=columns)
                    trigs.meta['segments'] = SegmentList([segment])
                else:
                    trigs = read_cache(segcache, SegmentList([segment]), etg,
                                       nproc=nproc, **read_kw)
                # record triggers
                if trigs is not None:
                    # add metadata
                    add_triggers(trigs, key)
                    ntrigs += len(trigs)
                vprint(".")
        vprint(" | %d events read\n" % ntrigs)

    # if asked to read triggers, but didn't actually read any,
    # create an empty table so that subsequent calls don't raise KeyErrors
    if query and key not in globalv.TRIGGERS:
        # find LIGO_LW table for this ETG
        try:
            if columns is not None:  # don't need to map to LIGO_LW
                raise KeyError
            TableClass = get_etg_table(etg)
        except KeyError:  # build simple table
            tab = EventTable(names=columns)
        else:  # map to LIGO_LW table with full column listing
            tab = EventTable(lsctables.New(TableClass))
        tab.meta['segments'] = SegmentList()
        for metakey in ('timecolumn', 'tablename',):
            if metakey in read_kw:
                tab.meta[metakey] = read_kw[metakey]
        add_triggers(tab, key)

    # work out time function
    if return_:
        return keep_in_segments(globalv.TRIGGERS[key], segments, etg)


def add_triggers(table, key, segments=None):
    """Add a `EventTable` to the global memory cache
    """
    if segments is not None:
        table.meta['segments'] = segments
    try:
        old = globalv.TRIGGERS[key]
    except KeyError:
        new = globalv.TRIGGERS[key] = table
        new.meta.setdefault('segments', SegmentList())
    else:
        new = globalv.TRIGGERS[key] = vstack_tables((old, table))
        new.meta = old.meta
        new.meta['segments'] |= table.meta.get('segments', SegmentList())
    new.meta['segments'].coalesce()
    return new


def keep_in_segments(table, segmentlist, etg=None):
    """Return a view of the table containing only those rows in the segmentlist
    """
    times = get_times(table, etg)
    keep = table_filters.in_segmentlist(times, segmentlist)
    out = table[keep]
    out.meta['segments'] = segmentlist & table.meta['segments']
    return out


def get_times(table, etg):
    """Get the time data for this table

    See Also
    --------
    get_time_column
    """
    try:
        return table[get_time_column(table, etg)]
    except ValueError:
        # match well-defined LIGO_LW table types
        tablename = table.meta.get('tablename') or ''
        for ttype, tcol in (
                ('_burst', 'peak'),
                ('_inspiral', 'end'),
                ('_ringdown', 'start'),
        ):
            if tablename.endswith(ttype):
                sec = '{0}_time'.format(tcol)
                nanosec = '{0}_time_ns'.format(tcol)
                return table[sec] + table[nanosec] * 1e-9
        raise


def get_time_column(table, etg):
    """Get the time column name for this table
    """
    # allow user to have selected the time column
    if table.meta.get('timecolumn'):
        return table.meta['timecolumn']
    # otherwise search for it
    try:
        return table._get_time_column()
    except ValueError:
        # shortcut pycbc
        if etg == 'pycbc_live':
            'end_time'

        # match well-defined LIGO_LW table types
        tablename = table.meta.get('tablename') or ''
        for ttype, tcol in (
                ('_burst', 'peak'),
                ('_inspiral', 'end'),
                ('_ringdown', 'start'),
        ):
            if tablename.endswith(ttype) and tcol in table.columns:
                table.meta['timecolumn'] = tcol
                return tcol
        raise


def get_etg_read_kwargs(etg, config=None, exclude=['columns']):
    """Read keyword arguments to pass to the trigger reader for a given etg
    """
    # use global defaults
    kwargs = {
        'format': None,
        'selection': [],
    }

    # get kwargs from config
    if config is not None and config.has_section(etg):
        config_kw = dict(config.nditems(etg))
    elif config is not None and config.has_section(etg.lower()):
        config_kw = dict(config.nditems(etg.lower()))
    else:
        config_kw = {}
    usrfmt = config_kw.get('format', None)

    # get ETG defaults : only if user didn't specify the read format,
    #                    or the format they did specify matches our default
    etgl = re_cchar.sub('_', etg).lower()
    if etgl in ETG_READ_KW and usrfmt in (None, ETG_READ_KW[etgl]['format']):
        kwargs.update(ETG_READ_KW.get(etgl, {}))

    # now add the config kwargs (so they override our defaults)
    kwargs.update(config_kw)

    # format kwargs
    for key in list(kwargs):
        # remove unwanted keys
        if key in exclude:
            kwargs.pop(key)
            continue
        # eval string to object (safely)
        kwargs[key] = safe_eval(kwargs[key])
        if key.endswith('columns') and isinstance(kwargs[key], str):
            kwargs[key] = kwargs[key].replace(' ', '').split(',')

    if 'selection' in kwargs:
        kwargs['selection'] = parse_column_filters(kwargs['selection'])

    return kwargs


def read_cache(cache, segments, etg, nproc=1, timecolumn=None, **kwargs):
    """Read a table of events from a cache

    This function is mainly meant for use from the `get_triggers` method

    Parameters
    ----------
    cache : :class:`glue.lal.Cache`
        the formatted list of files to read
    segments : `~gwpy.segments.SegmentList`
        the list of segments to read
    etg : `str`
        the name of the trigger generator that created the files
    nproc : `int`, optional
        the number of parallel processes to use when reading
    **kwargs
        other keyword arguments are passed to the `EventTable.read` or
        `{tableclass}.read` methods

    Returns
    -------
    table : `~gwpy.table.EventTable`, `None`
        a table of events, or `None` if the cache has no overlap with
        the segments
    """
    if isinstance(cache, Cache):
        cache = cache.sieve(segmentlist=segments)
        cache = cache.checkfilesexist()[0]
        cache.sort(key=lambda x: x.segment[0])
        cache = cache.pfnlist()  # some readers only like filenames
    else:
        cache = [urlparse(url).path for url in cache]
    if etg == 'pycbc_live':  # remove empty HDF5 files
        cache = filter_pycbc_live_files(cache, ifo=kwargs['ifo'])

    # Currently the EventTable.read() method fails if there are zero
    # length trigger files in addition to non-zero length trigger files.
    # Here, we read trigger files one-by-one and remove any with zero length
    # from the cache list.
    for idx, this_trigger_file in enumerate(cache):
        this_table = EventTable.read(this_trigger_file, **kwargs)
        if len(this_table) == 0:
            cache.remove(this_trigger_file)
            continue

    if len(cache) == 0:
        return

    # read triggers
    table = EventTable.read(cache, **kwargs)

    # store read keywords in the meta table
    if timecolumn:
        table.meta['timecolumn'] = timecolumn

    # get back from cache entry
    if isinstance(cache, CacheEntry):
        cache = Cache([cache])

    # append new events to existing table
    try:
        csegs = cache_segments(cache) & segments
    except (AttributeError, TypeError, ValueError):
        csegs = SegmentList()
    table.meta['segments'] = csegs

    if timecolumn:  # already filtered on-the-fly
        return table
    # filter now
    return keep_in_segments(table, segments, etg)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""A `Plot` is a representation of an image to be included in the HTML
output a :doc:`tab </tabs>`.

For simple purposes, a `Plot` is just a reference to an existing image file
that can be imported into an HTML page via the ``<img>`` tag.

For more complicated purposes, a number of data plot classes are provided to
allow users to generate images on-the-fly.
The available classes are:

.. autosummary::
   :toctree: api

   TimeSeriesDataPlot
   SpectrogramDataPlot
   SegmentDataPlot
   StateVectorDataPlot
   SpectrumDataPlot
   TimeSeriesHistogramPlot
   TriggerTimeSeriesDataPlot
   TriggerHistogramPlot
   TriggerRateDataPlot
"""

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

from .registry import (
    register_plot,
    get_plot,
)
from .utils import (
    get_column_label,
    get_column_string,
    hash,
)
from .core import (
    format_label,
    SummaryPlot,
    DataPlot,
    BarPlot,
    PiePlot,
)
from .builtin import (
    undo_demodulation,
    TimeSeriesDataPlot,
    SpectrogramDataPlot,
    CoherenceSpectrogramDataPlot,
    SpectrumDataPlot,
    CoherenceSpectrumDataPlot,
    TimeSeriesHistogramPlot,
    TimeSeriesHistogram2dDataPlot,
    SpectralVarianceDataPlot,
    RayleighSpectrogramDataPlot,
    RayleighSpectrumDataPlot,
)
from .segments import (
    tint_hex,
    common_limits,
    SegmentDataPlot,
    StateVectorDataPlot,
    DutyDataPlot,
    ODCDataPlot,
    SegmentPiePlot,
    NetworkDutyPiePlot,
    SegmentBarPlot,
    SegmentHistogramPlot,
)
from .triggers import (
    TriggerPlotMixin,
    TriggerDataPlot,
    TriggerTimeSeriesDataPlot,
    TriggerHistogramPlot,
    TriggerRateDataPlot,
)
from .range import (
    _get_params,
    RangePlotMixin,
    RangeDataPlot,
    RangeDataHistogramPlot,
    RangeSpectrogramDataPlot,
    RangeSpectrumDataPlot,
    RangeCumulativeSpectrumDataPlot,
    SimpleTimeVolumeDataPlot,
    GWpyTimeVolumeDataPlot,
)
from .noisebudget import (
    NoiseBudgetPlot,
    RelativeNoiseBudgetPlot,
)
from .guardian import GuardianStatePlot
from .sei import SeiWatchDogPlot



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013-2016)
#
# This file is part of GWSumm
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>

"""This module defines some utility `Tab` subclasses, including HTTP
error handlers.
"""

from MarkupPy import markup

from .registry import (get_tab, register_tab)

from gwdetchar.io import html

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__all__ = ['AboutTab', 'Error404Tab']

Tab = get_tab('basic')


# -- About --------------------------------------------------------------------

class AboutTab(Tab):
    """Page describing how the containing HTML pages were generated
    """
    type = 'about'

    def __init__(self, name='About', **kwargs):
        super(AboutTab, self).__init__(name, **kwargs)

    def write_html(self, config=list(), prog=None, **kwargs):
        return super(AboutTab, self).write_html(
            html.about_this_page(config=config, prog=prog), **kwargs)


register_tab(AboutTab)


# -- HTTP errors --------------------------------------------------------------

class Error404Tab(Tab):
    """Custom HTTP 404 error page
    """
    type = '404'

    def __init__(self, name='404', **kwargs):
        super(Error404Tab, self).__init__(name, **kwargs)

    def write_html(self, config=list(), top=None, **kwargs):
        if top is None:
            top = kwargs.get('base', self.path)
        kwargs.setdefault('title', '404: Page not found')
        page = markup.page()
        page.div(class_='alert alert-danger text-justify shadow-sm')
        page.p()
        page.strong("The page you are looking for does not exist.")
        page.p.close()
        page.p("This could be because the times for which you are looking "
               "were never processed (or have not happened yet), or because "
               "no page exists for the specific data products you want. "
               "Either way, if you think this is in error, please contact "
               "<a class=\"alert-link\" "
               "href=\"mailto:detchar+code@ligo.org\">the DetChar group</a>.")
        page.p("Otherwise, you might be interested in one of the following:")
        page.div(style="padding-top: 10px;")
        page.a("Take me back", role="button", class_="btn btn-lg btn-info",
               title="Back", href="javascript:history.back()")
        page.a("Take me up one level", role="button",
               class_="btn btn-lg btn-warning", title="Up",
               href="javascript:linkUp()")
        page.a("Take me to the top level", role="button",
               class_="btn btn-lg btn-success", title="Top", href=top)
        page.div.close()
        page.div.close()  # alert alert-danger
        page.script("""
  function linkUp() {
    var url = window.location.href;
    if (url.substr(-1) == '/') url = url.substr(0, url.length - 2);
    url = url.split('/');
    url.pop();
    window.location = url.join('/');
  }""", type="text/javascript")
        return super(Error404Tab, self).write_html(page, **kwargs)


register_tab(Error404Tab)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""This module handles HDF archiving of data.

In production for LIGO, the LIGO Summary Pages (LSP) Service runs at regular
intervals (about every 10 minutes), so an HDF5 file is used to archive the
data read and produced from one instance so that the next instance doesn't
have to re-read and re-produce the same data.

All data products are stored just using the 'standard' gwpy `.write()` method
for that object.
"""

import tempfile
import shutil
import warnings
import re
import datetime
import os

from numpy import (unicode_, ndarray)
from h5py import File

from astropy.table import Table

from gwpy.time import (from_gps, to_gps)
from gwpy.timeseries import (StateVector, TimeSeries)
from gwpy.spectrogram import Spectrogram
from gwpy.segments import (SegmentList, Segment, DataQualityFlag)

from . import (globalv, mode)
from .channels import get_channel
from .data import (add_timeseries, add_spectrogram,
                   add_coherence_component_spectrogram)
from .triggers import (EventTable, add_triggers)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

re_rate = re.compile('_EVENT_RATE_')


def write_data_archive(outfile, channels=True, timeseries=True,
                       spectrogram=True, segments=True, triggers=True):
    """Build and save an HDF archive of data processed in this job.

    Parameters
    ----------
    outfile : `str`
        path to target HDF5 file

    timeseries : `bool`, optional
        include `TimeSeries` data in archive

    spectrogram : `bool`, optional
        include `Spectrogram` data in archive

    segments : `bool`, optional
        include `DataQualityFlag` data in archive

    triggers : `bool`, optional
        include `EventTable` data in archive
    """

    # backup existing file, in case something fails
    backup = backup_existing_archive(outfile)

    try:
        with File(outfile, 'w') as h5file:

            # -- channels -----------------------

            if channels and globalv.CHANNELS:
                cols = ('name', 'sample_rate', 'frametype', 'unit')
                rows = []
                for chan in globalv.CHANNELS:
                    rows.append((
                        chan.ndsname,
                        chan.sample_rate.to('Hz').value if
                        chan.sample_rate is not None else 0,
                        str(getattr(chan, 'frametype', None)) or '',
                        str(chan.unit) if chan.unit else '',
                    ))
                Table(names=cols, rows=rows).write(h5file, 'channels')

            # -- timeseries ---------------------

            if timeseries:
                tgroup = h5file.create_group('timeseries')
                sgroup = h5file.create_group('statevector')
                # loop over channels
                for c, tslist in globalv.DATA.items():
                    c = get_channel(c)
                    # ignore trigger rate TimeSeries
                    if re_rate.search(str(c)):
                        continue
                    # loop over time-series
                    for ts in tslist:
                        # ignore fast channels who weren't used
                        # for a timeseries:
                        if (not isinstance(ts, StateVector) and
                                ts.sample_rate.value > 16.01 and
                                not getattr(c, '_timeseries', False)):
                            continue
                        # archive timeseries
                        try:
                            name = '%s,%s,%s' % (ts.name, ts.channel.ndsname,
                                                 ts.t0.value)
                        except AttributeError:
                            name = '%s,%s' % (ts.name, ts.t0.value)
                        if isinstance(ts, StateVector):
                            group = sgroup
                        else:
                            group = tgroup
                        _write_object(ts, group, path=name, format='hdf5')

            # -- spectrogram --------------------

            if spectrogram:
                for tag, gdict in zip(
                        ['spectrogram', 'coherence-components'],
                        [globalv.SPECTROGRAMS, globalv.COHERENCE_COMPONENTS]):
                    group = h5file.create_group(tag)
                    # loop over channels
                    for key, speclist in gdict.items():
                        # loop over time-series
                        for spec in speclist:
                            name = '%s,%s' % (key, spec.t0.value)
                            _write_object(spec, group, path=name,
                                          format='hdf5')

            # -- segments -----------------------

            if segments:
                group = h5file.create_group('segments')
                # loop over channels
                for name, dqflag in globalv.SEGMENTS.items():
                    dqflag.write(group, path=name, format='hdf5')

            # -- triggers -----------------------

            if triggers:
                group = h5file.create_group('triggers')
                for key in globalv.TRIGGERS:
                    archive_table(globalv.TRIGGERS[key], key, group)

    except Exception:  # if it fails for any reason, reinstate the backup
        if backup:
            restore_backup(backup, outfile)
        raise
    finally:  # remove the backup regardless of what happens
        if backup is not None and os.path.isfile(backup):
            os.remove(backup)


def read_data_archive(sourcefile, rm_source_on_fail=True):
    """Read archived data from an HDF5 archive source

    This method reads all found data into the data containers defined by
    the `gwsumm.globalv` module, then returns nothing.

    Parameters
    ----------
    sourcefile : `str`
        path to source HDF5 file

    rm_source_on_fail : `bool`, optional
        remove the source HDF5 file if there was an OSError when opening the
        file
    """
    # Check that the HDF5 source file is able to be opened.
    # A common failure mode is that the HDF5 file is corrupted and this brings
    # down the whole workflow, requiring manual intervention. Here, we attempt
    # to automatically catch a common failure
    try:
        h5file = File(sourcefile, 'r')
    except FileNotFoundError:
        raise
    except OSError as exc:  # file is corrupt, so we remove it to start fresh
        if not rm_source_on_fail:
            raise
        warnings.warn(f"failed to read {sourcefile} [{exc}], removing...")
        os.remove(sourcefile)
        return

    with File(sourcefile, 'r') as h5file:

        # -- channels ---------------------------

        try:
            ctable = Table.read(h5file['channels'])
        except KeyError:  # no channels table written
            pass
        else:
            for row in ctable:
                chan = get_channel(row['name'])
                for p in ctable.colnames[1:]:
                    if row[p]:
                        setattr(chan, p, row[p])

        # -- timeseries -------------------------

        for dataset in h5file.get('timeseries', {}).values():
            ts = TimeSeries.read(dataset, format='hdf5')
            if (re.search(r'\.(rms|min|mean|max|n)\Z', ts.channel.name) and
                    ts.sample_rate.value == 1.0):
                ts.channel.type = 's-trend'
            elif re.search(r'\.(rms|min|mean|max|n)\Z', ts.channel.name):
                ts.channel.type = 'm-trend'
            ts.channel = get_channel(ts.channel)
            try:
                add_timeseries(ts, key=ts.channel.ndsname)
            except ValueError:
                if mode.get_mode() != mode.Mode.day:
                    raise
                warnings.warn('Caught ValueError in combining daily archives')
                # get end time
                globalv.DATA[ts.channel.ndsname].pop(-1)
                t = globalv.DATA[ts.channel.ndsname][-1].span[-1]
                add_timeseries(ts.crop(start=t), key=ts.channel.ndsname)

        # -- statevector -- ---------------------

        for dataset in h5file.get('statevector', {}).values():
            sv = StateVector.read(dataset, format='hdf5')
            sv.channel = get_channel(sv.channel)
            add_timeseries(sv, key=sv.channel.ndsname)

        # -- spectrogram ------------------------

        for tag, add_ in zip(
                ['spectrogram', 'coherence-components'],
                [add_spectrogram, add_coherence_component_spectrogram]):
            for key, dataset in h5file.get(tag, {}).items():
                key = key.rsplit(',', 1)[0]
                spec = Spectrogram.read(dataset, format='hdf5')
                spec.channel = get_channel(spec.channel)
                add_(spec, key=key)

        # -- segments ---------------------------

        for name, dataset in h5file.get('segments', {}).items():
            dqflag = DataQualityFlag.read(h5file, path=dataset.name,
                                          format='hdf5')
            globalv.SEGMENTS += {name: dqflag}

        # -- triggers ---------------------------

        for dataset in h5file.get('triggers', {}).values():
            load_table(dataset)


def backup_existing_archive(filename, suffix='.h5',
                            prefix='gw_summary_archive_', dir=None):
    """Create a copy of an existing archive.
    """
    backup = tempfile.mktemp(suffix=suffix, prefix=prefix, dir=dir)
    try:
        shutil.move(filename, backup)
    except IOError:
        return None
    else:
        return backup


def restore_backup(backup, target):
    """Reinstate a backup copy of the archive.
    """
    shutil.move(backup, target)


def find_daily_archives(start, end, ifo, tag, basedir=os.curdir):
    """Find the daily archives spanning the given GPS [start, end) interval
    """
    archives = []
    s = from_gps(to_gps(start))
    e = from_gps(to_gps(end))
    while s < e:
        daybase = mode.get_base(s, mode=mode.Mode.day)
        ds = to_gps(s)
        s += datetime.timedelta(days=1)
        de = to_gps(s)
        archivedir = os.path.join(basedir, daybase, 'archive')
        arch = os.path.join(archivedir, '%s-%s-%d-%d.h5'
                            % (ifo, tag, ds, de-ds))
        if os.path.isfile(arch):
            archives.append(arch)
    return archives


# -- utility methods --------------------------------------------------------

def _write_object(data, *args, **kwargs):
    """Internal method to write something to HDF5 with error handling
    """
    try:
        return data.write(*args, **kwargs)
    except (TypeError, ValueError) as e:
        warnings.warn(str(e))
    except RuntimeError as e:
        if 'name already exists' in str(e).lower():
            warnings.warn(str(e))
        else:
            raise


def segments_to_array(segmentlist):
    """Convert a `SegmentList` to a 2-dimensional `numpy.ndarray`
    """
    out = ndarray((len(segmentlist), 2), dtype=float)
    for i, seg in enumerate(segmentlist):
        out[i] = seg
    return out


def segments_from_array(array):
    """Convert a 2-dimensional `numpy.ndarray` to a `SegmentList`
    """
    out = SegmentList()
    for row in array:
        out.append(Segment(*row))
    return out


def archive_table(table, key, parent):
    """Add a table to the given HDF5 group

    .. warning::

       If the input ``table`` is empty, it will not be archived

    Parameters
    ----------
    table : `~astropy.table.Table`
        the data to archive

    key : `str`
        the path (relative to ``parent``) at which to store the table

    parent : `h5py.Group`
        the h5py group in which to add this dataset

    """
    # copy table so we can modify meta without impacting anything
    table = table.copy(copy_data=False)

    if len(table) == 0:
        warnings.warn("%r table is empty and will not be archived" % key)
        return
    table.meta.pop('psd', None)  # pycbc_live
    table.meta.pop('loudest', None)  # pycbc_live
    try:
        table.meta['segments'] = segments_to_array(table.meta['segments'])
    except KeyError:
        pass
    for col in table.columns:
        if table[col].dtype.type is unicode_:
            table.replace_column(col, table[col].astype(str))
    table.write(parent, path=key, format='hdf5')
    return key


def load_table(dataset):
    """Read table from the given HDF5 group

    The `EventTable` is read, stored in the memory archive, then returned

    Parameters
    ----------
    dataset : `h5py.Dataset`
        n-dimensional table to load from hdf5

    Returns
    -------
    table : `~gwpy.table.EventTable`
        the table of events loaded from hdf5
    """
    table = EventTable.read(dataset, format='hdf5')
    try:
        table.meta['segments'] = segments_from_array(table.meta['segments'])
    except KeyError:
        table.meta['segments'] = SegmentList()
    add_triggers(table, dataset.name.split('/')[-1])
    return table



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwsumm.archive`

"""

import os
import tempfile

import pytest

import h5py

from numpy import (random, testing as nptest)

from gwpy.table import EventTable
from gwpy.timeseries import (TimeSeries, StateVector)
from gwpy.spectrogram import Spectrogram
from gwpy.segments import (Segment, SegmentList)

from .. import (archive, data, globalv, channels, triggers)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

TEST_DATA = TimeSeries([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], epoch=100,
                       unit='meter', sample_rate=1, channel='X1:TEST-CHANNEL',
                       name='TEST DATA')
TEST_DATA.channel = channels.get_channel(TEST_DATA.channel)


# -- utilities ----------------------------------------------------------------

def empty_globalv():
    globalv.DATA = type(globalv.DATA)()
    globalv.SPECTROGRAMS = type(globalv.SPECTROGRAMS)()
    globalv.SEGMENTS = type(globalv.SEGMENTS)()
    globalv.TRIGGERS = type(globalv.TRIGGERS)()


def create(data, **metadata):
    SeriesClass = metadata.pop('series_class', TimeSeries)
    d = SeriesClass(data, **metadata)
    d.channel = channels.get_channel(d.channel)
    if not d.name:
        d.name = d.channel.texname
    return d


# -- tests --------------------------------------------------------------------

def test_write_archive(delete=True):
    empty_globalv()
    data.add_timeseries(TEST_DATA)
    data.add_timeseries(create([1, 2, 3, 4, 5],
                               dt=60., channel='X1:TEST-TREND.mean'))
    data.add_timeseries(create([1, 2, 3, 2, 1],
                               series_class=StateVector,
                               channel='X1:TEST-STATE_VECTOR'))
    data.add_spectrogram(create([[1, 2, 3], [3, 2, 1], [1, 2, 3]],
                                series_class=Spectrogram,
                                channel='X1:TEST-SPECTROGRAM'))
    t = EventTable(random.random((100, 5)), names=['time', 'a', 'b', 'c', 'd'])
    t.meta['segments'] = SegmentList([Segment(0, 100)])
    triggers.add_triggers(t, 'X1:TEST-TABLE,testing')
    fname = tempfile.mktemp(suffix='.h5', prefix='gwsumm-tests-')
    try:
        archive.write_data_archive(fname)
        archive.write_data_archive(fname)  # test again to validate backups
    finally:
        if delete and os.path.isfile(fname):
            os.remove(fname)
    return fname


def test_read_archive():
    fname = test_write_archive(delete=False)
    empty_globalv()
    try:
        archive.read_data_archive(fname)
    finally:
        os.remove(fname)
    # check timeseries
    ts = data.get_timeseries('X1:TEST-CHANNEL',
                             [(100, 110)], query=False).join()
    nptest.assert_array_equal(ts.value, TEST_DATA.value)
    for attr in ['epoch', 'unit', 'sample_rate', 'channel', 'name']:
        assert getattr(ts, attr) == getattr(TEST_DATA, attr)
    # check trend series
    ts = data.get_timeseries('X1:TEST-TREND.mean,m-trend', [(0, 300)],
                             query=False).join()
    assert ts.channel.type == 'm-trend'
    assert ts.span == (0, 300)
    # check triggers
    t = triggers.get_triggers('X1:TEST-TABLE', 'testing', [(0, 100)])
    assert len(t) == 100


def test_archive_load_table():
    t = EventTable(random.random((100, 5)),
                   names=['a', 'b', 'c', 'd', 'e'])
    empty = EventTable(names=['a', 'b'])
    try:
        fname = tempfile.mktemp(suffix='.h5', prefix='gwsumm-tests-')
        h5file = h5py.File(fname, mode='a')
        # check table gets archived and read transparently
        archive.archive_table(t, 'test-table', h5file)
        t2 = archive.load_table(h5file['test-table'])
        nptest.assert_array_equal(t.as_array(), t2.as_array())
        assert t.dtype == t2.dtype
        # check empty table does not get archived, with warning
        with pytest.warns(UserWarning):
            n = archive.archive_table(empty, 'test-empty', h5file)
        assert n is None
        assert 'test-empty' not in h5file
    finally:
        if os.path.exists(fname):
            os.remove(fname)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2017)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output utilities
"""

import os.path
import re

from astropy.io.registry import IORegistryError

from gwpy.frequencyseries import FrequencySeries

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

HDF5_FILENAME = re.compile(r'(?P<ext>(.hdf5|.hdf|.h5))\/')


def read_frequencyseries(filename):
    """Read a `~gwpy.frequencyseries.FrequencySeries` from a file

    IF using HDF5, the filename can be given as a combined filename/path, i.e.
    ``test.hdf5/path/to/dataset``.

    Parameters
    ----------
    filename : `str`
        path of file to read

    Returns
    -------
    series : `~gwpy.frequencyseries.FrequencySeries`
        the data as read

    Raises
    ------
    astropy.io.registry.IORegistryError
        if the input format cannot be identified or is not registered
    """
    # try and parse path in HDF5 file if given
    try:
        ext = HDF5_FILENAME.search(filename).groupdict()['ext']
    except AttributeError:  # no match
        kwargs = {}
    else:
        kwargs = {'path': filename.rsplit(ext, 1)[1]}
    # read file
    try:
        return FrequencySeries.read(filename, **kwargs)
    except IORegistryError:
        if filename.endswith('.gz'):
            fmt = os.path.splitext(filename[:-3])[-1]
        else:
            fmt = os.path.splitext(filename)[-1]
        return FrequencySeries.read(filename, format=fmt.lstrip('.'), **kwargs)



# coding=utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Command-line interface for the gravitational-wave interferometer summary
information system (`gwsumm`).

This module provides the command-line interface to the GWSumm package,
allowing generation of detector summary information.

Select a <mode> to run over a calendar amount of time ('day', 'week',
or 'month'), or an arbitrary GPS (semi-open) interval.

Run '%(prog)s <mode> --help' for details of the specific arguments and
options available for each mode.
"""

import argparse
import calendar
import datetime
import getpass
import os
import re
import sys
import warnings

from collections import OrderedDict
from configparser import (DEFAULTSECT, NoOptionError, NoSectionError)
from dateutil.relativedelta import relativedelta
from urllib.parse import urlparse

from glue.lal import (Cache, LIGOTimeGPS)

from gwpy import time
from gwpy.segments import (Segment, SegmentList)
from gwpy.signal.spectral import _lal as fft_lal
from gwpy.time import (Time, _tconvert, tconvert, to_gps)

from gwdetchar.cli import logger

from . import (
    __version__,
    archive,
    globalv,
    mode,
)
from .config import (
    GWSummConfigParser,
)
from .segments import get_segments
from .state import (
    ALLSTATE
)
from .tabs import (
    TabList,
    get_tab,
)
from .utils import (
    get_default_ifo,
    mkdir,
    re_flagdiv,
)
from .data import get_timeseries_dict

# set matplotlib backend
from matplotlib import use
use('Agg')

# XXX HACK XXX
# plots don't work with multiprocessing with lal.LIGOTimeGPS
time.LIGOTimeGPS = LIGOTimeGPS
_tconvert.LIGOTimeGPS = LIGOTimeGPS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

# set defaults
VERBOSE = False
PROFILE = False
try:
    DEFAULT_IFO = get_default_ifo()
except ValueError:
    DEFAULT_IFO = None

# initialize logger
PROG = ('python -m gwsumm' if sys.argv[0].endswith('.py')
        else os.path.basename(sys.argv[0]))
LOGGER = logger(name=PROG.split('python -m ').pop())

# find today's date
TODAY = datetime.datetime.utcnow().strftime('%Y%m%d')


# -- argparse utilities -------------------------------------------------------

class GWArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super(GWArgumentParser, self).__init__(*args, **kwargs)
        self._positionals.title = 'Positional arguments'
        self._optionals.title = 'Optional arguments'


class GWHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('indent_increment', 4)
        super(GWHelpFormatter, self).__init__(*args, **kwargs)


class DateAction(argparse.Action):
    TIMESCALE = {'days': 1}

    @staticmethod
    def set_gps_times(namespace, startdate, enddate):
        setattr(namespace, 'gpsstart', to_gps(startdate))
        setattr(namespace, 'gpsend', to_gps(enddate))

    def __call__(self, parser, namespace, values, option_string=None):
        try:
            date = datetime.datetime.strptime(values, self.DATEFORMAT)
        except ValueError:
            raise parser.error("%s malformed: %r. Please format as %s"
                               % (self.dest.title(), values, self.METAVAR))
        else:
            self.set_gps_times(namespace, date,
                               date + relativedelta(**self.TIMESCALE))
            setattr(namespace, self.dest, date)
        return date


class DayAction(DateAction):
    TIMESCALE = {'days': 1}
    DATEFORMAT = '%Y%m%d'
    METAVAR = 'YYYYMMDD'


class WeekAction(DayAction):
    TIMESCALE = {'days': 7}


class MonthAction(DateAction):
    TIMESCALE = {'months': 1}
    DATEFORMAT = '%Y%m'
    METAVAR = 'YYYYMM'


class YearAction(DateAction):
    TIMESCALE = {'years': 1}
    DATEFORMAT = '%Y'
    METAVAR = 'YYYY'


class GPSAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=False):
        try:
            values = float(values)
        except (TypeError, ValueError):
            pass
        setattr(namespace, self.dest, to_gps(values))


# -- parse command-line -------------------------------------------------------

def add_output_options(parser_):
    """Add outuput options to the subparser

    This is only needed because argparse can't handle mutually exclusive
    groups in a parent parser handed to a subparser for some reason.
    """
    outopts = parser_.add_argument_group("Output options")
    outopts.add_argument(
        '-o',
        '--output-dir',
        action='store',
        type=str,
        metavar='DIR',
        default=os.curdir,
        help="Output directory for summary information",
    )
    htmlopts = outopts.add_mutually_exclusive_group()
    htmlopts.add_argument(
        '-m',
        '--html-only',
        action='store_true',
        default=False,
        help="Generate container HTML and navigation only",
    )
    htmlopts.add_argument(
        '-n',
        '--no-html',
        action='store_true',
        default=False,
        help="Generate inner HTML and contents only, not supporting HTML",
    )
    outopts.add_argument(
        '-N',
        '--no-htaccess',
        action='store_true',
        default=False,
        help="don't create a .htaccess file to customise 404 errors",
    )


def add_archive_options(parser_):
    """Add archiving options to the subparser

    This is only needed because argparse can't handle mutually exclusive
    groups in a parent parser handed to a subparser for some reason.
    """
    hierarchopts = parser_.add_argument_group('Archive options')
    hierarchchoice = hierarchopts.add_mutually_exclusive_group()
    hierarchchoice.add_argument(
        '-a',
        '--archive',
        metavar='FILE_TAG',
        default=False,
        const='GW_SUMMARY_ARCHIVE',
        nargs='?',
        help="Read archived data from, and write processed data to "
             "an HDF archive file written with the FILE_TAG. If not "
             "given, no archive will be used, if given with no file "
             "tag, a default of '%(const)s' will be used.",
    )
    hierarchchoice.add_argument(
        '-d',
        '--daily-archive',
        metavar='FILE_TAG',
        default=False,
        const='GW_SUMMARY_ARCHIVE',
        nargs='?',
        help="Read data from the daily archives, with the given FILE_TAG."
             "If not given, daily archives will be used, if given with no "
             "file tag, a default of '%(const)s' will be used.",
    )


def create_parser():
    """Create a command-line parser for this entry point
    """
    # initialize top-level argument parser
    parser = GWArgumentParser(
        formatter_class=GWHelpFormatter,
        prog=PROG,
        description=__doc__,
        fromfile_prefix_chars="@",
        epilog="Arguments and options may be written into files and passed as "
               "positional arguments prepended with '@', e.g. '%(prog)s "
               "@args.txt'. In this format, options must be give as "
               "'--argument=value', and not '--argument value'.",
    )

    # global arguments
    parser.add_argument(
        '-V',
        '--version',
        action='version',
        version=__version__,
        help="show program's version number and exit",
    )

    # shared arguments
    sharedopts = GWArgumentParser(add_help=False)
    sharedopts.title = 'Progress arguments'
    sharedopts.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        default=False,
        help="show verbose logging output",
    )
    sharedopts.add_argument(
        '-D',
        '--debug',
        action='store_true',
        default=False,
        help="show information that could be useful in debugging",
    )

    # configuration arguments
    copts = sharedopts.add_argument_group(
        "Configuration options",
        "Provide a number of INI-format configuration files",
    )
    copts.add_argument(
        '-i',
        '--ifo',
        default=DEFAULT_IFO,
        metavar='IFO',
        help="IFO prefix for interferometer to process. "
             "If this option is set in the [DEFAULT] of any of "
             "the INI files, giving it here is redundant.",
    )
    copts.add_argument(
        '-f',
        '--config-file',
        action='append',
        type=str,
        metavar='FILE',
        default=[],
        help="INI file for analysis, may be given multiple times",
    )
    copts.add_argument(
        '-t',
        '--process-tab',
        action='append',
        type=str,
        help="process only this tab, can be given multiple times",
    )

    # process options
    popts = sharedopts.add_argument_group(
        "Process options",
        "Configure how this summary will be processed.",
    )
    popts.add_argument(
        '--nds',
        action='store_true',
        default=None,
        help="use NDS as the data source, default: 'guess'",
    )
    popts.add_argument(
        '-j',
        '--multi-process',
        action='store',
        type=int,
        default=1,
        dest='multiprocess',
        metavar='N',
        help="use a maximum of N parallel processes at any time",
    )
    popts.add_argument(
        '-b',
        '--bulk-read',
        action='store_true',
        default=False,
        help="read all data up-front at the start of the job, "
             "rather than when it is needed for a tab",
    )
    popts.add_argument(
        '-S',
        '--on-segdb-error',
        action='store',
        type=str,
        default='raise',
        choices=['raise', 'ignore', 'warn'],
        help="action upon error fetching segments from SegDB",
    )
    popts.add_argument(
        '-G',
        '--on-datafind-error',
        action='store',
        type=str,
        default='raise',
        choices=['raise', 'ignore', 'warn'],
        help="action upon error querying for frames from the "
             "datafind server, default: %(default)s",
    )
    popts.add_argument(
        '--data-cache',
        action='append',
        default=[],
        help='path to LAL-format cache of TimeSeries data files',
    )
    popts.add_argument(
        '--event-cache',
        action='append',
        default=[],
        help='path to LAL-format cache of event trigger files',
    )
    popts.add_argument(
        '--segment-cache',
        action='append',
        default=[],
        help='path to LAL-format cache of state '
             'or data-quality segment files',
    )

    # define sub-parser handler
    subparsers = parser.add_subparsers(
        dest='mode',
        title='Modes',
        description='Note: all dates are defined with UTC boundaries.\n'
                    'The valid modes are:',
    )
    subparser = dict()

    # DAY mode
    daydoc = """
    Run %s over a full UTC day, and link this day to others with a calendar
    built into the HTML navigation bar. In this mode you can also archive data
    in HDF-format files to allow progressive processing of live data without
    restarting from scratch every time.""" % parser.prog
    subparser['day'] = subparsers.add_parser(
        'day',
        description=daydoc,
        epilog=parser.epilog,
        parents=[sharedopts],
        formatter_class=GWHelpFormatter,
        help="Process one day of data",
    )
    subparser['day'].add_argument(
        'day',
        action=DayAction,
        type=str,
        nargs='?',
        metavar=DayAction.METAVAR,
        default=TODAY,
        help="Day to process",
    )
    add_output_options(subparser['day'])

    darchopts = subparser['day'].add_argument_group(
        'Archive options',
        'Choose if, and how, to archive data from this run',
    )
    darchopts.add_argument(
        '-a',
        '--archive',
        metavar='FILE_TAG',
        default=False,
        const='GW_SUMMARY_ARCHIVE',
        nargs='?',
        help="Read archived data from, and write processed "
             "data to, an HDF archive file written with the "
             "FILE_TAG. If not given, no archive will be used, "
             "if given with no file tag, a default of "
             "'%(const)s' will be used.",
    )

    # WEEK mode
    subparser['week'] = subparsers.add_parser(
        'week',
        parents=[sharedopts],
        epilog=parser.epilog,
        formatter_class=GWHelpFormatter,
        help="Process one week of data",
    )
    subparser['week'].add_argument(
        'week',
        action=WeekAction,
        type=str,
        metavar=WeekAction.METAVAR,
        help="Week to process (given as starting day)",
    )
    add_output_options(subparser['week'])
    add_archive_options(subparser['week'])

    # MONTH mode
    subparser['month'] = subparsers.add_parser(
        'month',
        parents=[sharedopts],
        epilog=parser.epilog,
        formatter_class=GWHelpFormatter,
        help="Process one month of data",
    )
    subparser['month'].add_argument(
        'month',
        action=MonthAction,
        type=str,
        metavar=MonthAction.METAVAR,
        help="Month to process",
    )
    add_output_options(subparser['month'])
    add_archive_options(subparser['month'])

    # GPS mode
    subparser['gps'] = subparsers.add_parser(
        'gps',
        parents=[sharedopts],
        epilog=parser.epilog,
        formatter_class=GWHelpFormatter,
        help="Process GPS interval",
    )
    subparser['gps'].add_argument(
        'gpsstart',
        action=GPSAction,
        type=str,
        metavar='GPSSTART',
        help='GPS start time',
    )
    subparser['gps'].add_argument(
        'gpsend',
        action=GPSAction,
        type=str,
        metavar='GPSEND',
        help='GPS end time.',
    )

    garchopts = subparser['gps'].add_argument_group(
        'Archive options',
        'Choose if, and how, to archive data from this run',
    )
    garchopts.add_argument(
        '-a',
        '--archive',
        metavar='FILE_TAG',
        default=False,
        const='GW_SUMMARY_ARCHIVE',
        nargs='?',
        help="Read archived data from, and write processed "
             "data to, an HDF archive file written with the "
             "FILE_TAG. If not given, no archive will be used, "
             "if given with no file tag, a default of "
             "'%(const)s' will be used.")

    add_output_options(subparser['gps'])

    # return the argument parser
    return parser


# -- main code block ----------------------------------------------------------

def main(args=None):
    """Run the GWSumm command-line interface
    """
    parser = create_parser()
    args = parser.parse_args(args=args)

    if args.debug:
        warnings.simplefilter('error', DeprecationWarning)

    # set verbose output options
    globalv.VERBOSE = args.verbose

    # find all config files
    args.config_file = [os.path.expanduser(fp) for csv in args.config_file for
                        fp in csv.split(',')]

    # check segdb option
    if args.on_segdb_error not in ['raise', 'warn', 'ignore']:
        parser.error("Invalid option --on-segdb-error='%s'" %
                     args.on_segdb_error)

    # read configuration file
    config = GWSummConfigParser()
    config.optionxform = str
    if args.ifo:
        config.set_ifo_options(args.ifo, section=DEFAULTSECT)
    config.set(DEFAULTSECT, 'user', getpass.getuser())
    config.read(args.config_file)

    try:
        ifo = config.get(DEFAULTSECT, 'IFO')
    except NoOptionError:
        ifo = None
    finally:
        globalv.IFO = ifo

    # interpolate section names
    interp = {}
    if ifo:
        interp['ifo'] = ifo.lower()
        interp['IFO'] = ifo.title()
    config.interpolate_section_names(**interp)

    # double-check week mode matches calendar setting
    if args.mode == 'week':
        if config.has_option("calendar", "start-of-week"):
            weekday = getattr(calendar,
                              config.get("calendar", "start-of-week").upper())
            if weekday != args.week.timetuple().tm_wday:
                msg = ("Cannot process week starting on %s. The "
                       "'start-of-week' option in the [calendar] section "
                       "of the INI file specifies weeks start on %ss."
                       % (args.week.strftime('%Y%m%d'),
                          config.get("calendar", "start-of-week")))
                raise parser.error(msg)

    # record times in ConfigParser
    config.set_date_options(args.gpsstart, args.gpsend, section=DEFAULTSECT)

    # convert times for convenience
    span = Segment(args.gpsstart, args.gpsend)
    utc = tconvert(args.gpsstart)
    starttime = Time(float(args.gpsstart), format='gps')
    endtime = Time(float(args.gpsend), format='gps')

    # set mode and output directory
    # This will set the output directory to be day/YYYYMMDD, week/YYYYMMDD,
    # month/YYYYMM, or gps/<gps start>-<gps end>.
    mode.set_mode(args.mode)
    try:
        path = mode.get_base(utc)
    except ValueError:
        path = os.path.join('gps', f'{args.gpsstart}-{args.gpsend}')

    # set LAL FFT plan wisdom level
    duration = min(globalv.NOW, args.gpsend) - args.gpsstart
    if duration > 200000:
        fft_lal.LAL_FFTPLAN_LEVEL = 3
    elif duration > 40000:
        fft_lal.LAL_FFTPLAN_LEVEL = 2
    else:
        fft_lal.LAL_FFTPLAN_LEVEL = 1

    # set global html only flag
    if args.html_only:
        globalv.HTMLONLY = True

    # build directories
    mkdir(args.output_dir)
    os.chdir(args.output_dir)
    plotdir = os.path.join(path, 'plots')
    mkdir(plotdir)

    # -- setup --------------------------------------

    LOGGER.info(" -- GW interferometer summary information system -- ")
    LOGGER.debug("This is process {}".format(os.getpid()))
    LOGGER.debug("You have selected {} mode".format(mode.get_mode().name))
    LOGGER.debug("Start time: {0} ({1})".format(starttime.utc.iso,
                                                starttime.gps))
    LOGGER.debug("End time: {0} ({1})".format(endtime.utc.iso,
                                              endtime.gps))
    LOGGER.debug("Output directory: {}".format(
        os.path.abspath(os.path.join(args.output_dir, path))))

    # -- Finalise configuration
    LOGGER.info("Loading configuration")
    plugins = config.load_plugins()
    if plugins:
        LOGGER.debug(" -- Loaded {} plugins:".format(len(plugins)))
        for mod in plugins:
            LOGGER.debug("        %s" % mod)
    units = config.load_units()
    LOGGER.debug("    Loaded %d units" % len(units))
    channels = config.load_channels()
    LOGGER.debug("    Loaded %d channels" % len(channels))
    states = config.load_states()
    LOGGER.debug("    Loaded %d states" % len(states))
    rcp = config.load_rcParams()
    LOGGER.debug("    Loaded %d rcParams" % len(rcp))

    # read list of tabs
    tablist = TabList.from_ini(config, match=args.process_tab,
                               path=path, plotdir=plotdir)
    tablist.sort(reverse=True)
    tabs = sorted(tablist.get_hierarchy(), key=tablist._sortkey)
    LOGGER.info("    Loaded %d tabs [%d parents overall]" % (
        len(tablist), len(tabs)))

    # read caches
    cache = {}
    for (key, var) in zip(['datacache', 'trigcache', 'segmentcache'],
                          [args.data_cache, args.event_cache,
                           args.segment_cache]):
        if var:
            LOGGER.info("Reading %s from %d files... " % (key, len(var)))
            cache[key] = Cache()
            for fp in var:
                with open(fp, 'r') as f:
                    cache[key].extend(Cache.fromfile(f))
            cache[key] = cache[key].sieve(segment=span)
            LOGGER.debug("Done [%d entries]" % len(cache[key]))

    # -- read archive -------------------------------

    if not hasattr(args, 'archive'):
        args.archive = False

    if args.html_only:
        args.archive = False
        args.daily_archive = False
    elif args.archive is True:
        args.archive = 'GW_SUMMARY_ARCHIVE'

    archives = []

    if args.archive:
        archivedir = os.path.join(path, 'archive')
        mkdir(archivedir)
        args.archive = os.path.join(archivedir, '%s-%s-%d-%d.h5'
                                    % (ifo, args.archive, args.gpsstart,
                                       args.gpsend - args.gpsstart))
        if os.path.isfile(args.archive):
            archives.append(args.archive)
        else:
            LOGGER.debug(
                "No archive found in %s, one will be created at the end"
                % args.archive)

    # read daily archive for week/month/... mode
    if hasattr(args, 'daily_archive') and args.daily_archive:
        # find daily archive files
        archives.extend(archive.find_daily_archives(
            args.gpsstart, args.gpsend, ifo, args.daily_archive, archivedir))
        # then don't read any actual data
        cache['datacache'] = Cache()

    for arch in archives:
        LOGGER.info("Reading archived data from %s" % arch)
        archive.read_data_archive(arch)
        LOGGER.debug("Archive data loaded")

    # -- read HTML configuration --------------------

    css = config.get_css(section='html')
    javascript = config.get_javascript(section='html')

    # enable comments
    try:
        globalv.HTML_COMMENTS_NAME = config.get('html', 'disqus-shortname')
    except (NoOptionError, NoSectionError):
        pass

    # find new ifo bases
    ifobases = {}
    try:
        bases_ = config.nditems('html')
    except NoSectionError:
        pass
    else:
        base_reg = re.compile(r'-base\Z')
        for key, val in bases_:
            if base_reg.search(key):
                ifobases[key.rsplit('-', 1)[0]] = val
    ifobases = OrderedDict(sorted(ifobases.items(), key=lambda x: x[0]))

    # -- write auxiliary pages ----------------------

    # get URL from output directory
    if 'public_html' in os.getcwd():
        urlbase = os.path.sep + os.path.join(
                      '~%s' % config.get(DEFAULTSECT, 'user'),
                      os.getcwd().split('public_html', 1)[1][1:])
        base = urlbase
    # otherwise get URL from html config
    elif ifo in ifobases:
        urlbase = urlparse(ifobases[ifo]).path
        base = urlbase
    # otherwise let the write_html processor work it out on-the-fly
    else:
        urlbase = None
        base = None

    # get link to issues report page
    try:
        issues = config.get('html', 'issues')
    except (NoSectionError, KeyError):
        issues = True

    # write 404 error page
    if not args.no_htaccess and not args.no_html and urlbase:
        top = os.path.join(urlbase, path)
        four0four = get_tab('404')(span=span, parent=None, path=path,
                                   index=os.path.join(path, '404.html'))
        four0four.write_html(css=css, js=javascript, tabs=tabs, ifo=ifo,
                             ifomap=ifobases, top=top, base=base,
                             writedata=not args.html_only,
                             writehtml=not args.no_html,
                             issues=issues)
        url404 = os.path.join(urlbase, four0four.index)
        with open(os.path.join(path, '.htaccess'), 'w') as htaccess:
            print('Options -Indexes', file=htaccess)
            print('ErrorDocument 404 %s' % url404, file=htaccess)
            print('ErrorDocument 403 %s' % url404, file=htaccess)

    # write config page
    about = get_tab('about')(span=span, parent=None, path=path)
    if not args.no_html:
        mkdir(about.path)
        about.write_html(
            css=css, js=javascript, tabs=tabs, config=config.files,
            prog=PROG, ifo=ifo, ifomap=ifobases, about=about.index,
            base=base, issues=issues, writedata=not args.html_only,
            writehtml=not args.no_html)

    # -- read bulk data -----------------------------

    # XXX: bulk data reading could optimise things
    # XXX: but has never been used, so should remove (DMM 18/01/16)
    if args.bulk_read and not args.html_only:
        LOGGER.info("Reading all data in BULK")
        allts = set()
        allsv = set()
        allflags = set()
        for tab in tablist:
            snames = []
            for state in tab.states:
                snames.append(state.name)
                if state.definition:
                    allflags.update(re_flagdiv.split(state.definition))
            # get all data defined for the 'all' state
            if ALLSTATE in snames:
                allts.update(tab.get_channels('timeseries', 'spectrogram',
                                              'spectrum', 'histogram'))
                allsv.update(tab.get_channels('statevector'))
                allflags.update(tab.get_flags('segments'))
            # or get data for plots defined over all states
            else:
                for plot in tab.plots:
                    if plot.state is not None:
                        continue
                    if plot.type in ['timeseries', 'spectrogram', 'spectrum',
                                     'histogram']:
                        allts.update(plot.channels)
                    elif plot.type in ['statevector']:
                        allsv.update(plot.channels)
                    elif plot.type in ['segments']:
                        allflags.update([f for cflag in plot.flags for f in
                                         re_flagdiv.split(cflag)[::2] if f])
        allseg = SegmentList([span])
        if len(allflags):
            LOGGER.info(
                "%d data-quality flags identified for segment query from all "
                "tabs" % len(allflags))
            get_segments(allflags, allseg, config=config, return_=False)
        if len(allts):
            LOGGER.info("%d channels identified for TimeSeries from all tabs"
                        % len(allts))
            get_timeseries_dict(allts, allseg,
                                config=config, nds=args.nds,
                                nproc=args.multiprocess, return_=False)
        if len(allsv):
            LOGGER.info("%d channels identified for StateVector from all tabs"
                        % len(allsv))
            get_timeseries_dict(allsv, allseg,
                                config=config, nds=args.nds, statevector=True,
                                nproc=args.multiprocess, return_=False)

    # -- process all tabs ---------------------------

    # TODO: consider re-working this loop as TabList.process_all

    for tab in tablist:
        if tab.parent:
            name = '%s/%s' % (tab.parent.name, tab.name)
        else:
            name = tab.name
        if not args.html_only and isinstance(tab, get_tab('_processed')):
            LOGGER.debug("Processing %s" % name)
            tab.process(config=config, nds=args.nds,
                        nproc=args.multiprocess,
                        segdb_error=args.on_segdb_error,
                        datafind_error=args.on_datafind_error, **cache)
        if not tab.hidden and not isinstance(tab, get_tab('link')):
            mkdir(tab.href)
            tab.write_html(
                css=css, js=javascript, tabs=tabs, ifo=ifo,
                ifomap=ifobases, about=about.index, base=base,
                issues=issues, writedata=not args.html_only,
                writehtml=not args.no_html)

        # archive this tab
        if args.archive:
            LOGGER.info("Writing data to archive")
            archive.write_data_archive(args.archive)
            LOGGER.debug("Archive written to {}".format(
                os.path.abspath(args.archive)))
        LOGGER.debug("%s complete" % (name))

    LOGGER.info("-- Data products written, all done --")


# -- run from command-line ----------------------------------------------------

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2020)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwsumm.batch` command-line interface
"""

import os
import pytest
import shutil

from unittest import mock

from .. import batch

__author__ = 'Alex Urban <alexander.urban@ligo.org>'


# -- utilities ----------------------------------------------------------------

def _get_inputs():
    """Prepare and return paths to input data products
    """
    indir = os.getcwd()
    inputs = (
        os.path.join(indir, "global.ini"),
        os.path.join(indir, "k1-test.ini"),
        os.path.join(indir, "x509.cert"),
    )
    # write empty input files
    for filename in inputs:
        with open(filename, 'w') as f:
            f.write("")
    return inputs


# -- cli tests ----------------------------------------------------------------

@mock.patch(
    'gwsumm.batch.find_credential',
)
@mock.patch(
    'gwpy.io.kerberos.kinit',
    return_value=None,
)
def test_main(krb, x509, tmpdir, caplog):
    outdir = str(tmpdir)
    (global_, k1test, x509cert) = _get_inputs()
    x509.return_value = (x509cert, x509cert)
    args = [
        '--verbose',
        '--ifo', 'K1',
        '--maxjobs', '5',
        '--condor-timeout', '3',
        '--condor-command', 'notification=false',
        '--config-file', k1test,
        '--global-config', global_,
        '--nds',
        '--multi-process', '4',
        '--archive',
        '--event-cache', '/this/cache/is/not/real.cache',
        '--no-htaccess',
        '--output-dir', outdir,
    ]
    # test log output
    batch.main(args)
    assert "Copied all INI configuration files to ./etc" in caplog.text
    assert ("Configured Condor and Kerberos for NFS-shared credentials"
            in caplog.text)
    assert " -- Configured HTML htmlnode job" in caplog.text
    assert " -- Configured job for config {}".format(
        os.path.join(outdir, "etc", os.path.basename(k1test))) in caplog.text
    assert "Setup complete, DAG written to: {}".format(
        os.path.join(outdir, "gw_summary_pipe.dag")) in caplog.text
    # test file output
    assert set(os.listdir(outdir)) == {
        "etc",
        "gw_summary_pipe_local.sub",
        "gw_summary_pipe.dag",
        "logs",
        "gw_summary_pipe.sub",
        "gw_summary_pipe.sh",
    }
    assert set(os.listdir(os.path.join(outdir, "etc"))) == {
        os.path.basename(k1test),
        os.path.basename(x509cert),
        os.path.basename(global_),
    }
    assert set(os.listdir(os.path.join(outdir, "logs"))) == set()
    # clean up
    for filename in (global_, k1test, x509cert):
        os.remove(filename)
    shutil.rmtree(outdir, ignore_errors=True)


@pytest.mark.parametrize(
    'mode',
    (['--day', '20170209'],
     ['--week', '20170209'],
     ['--month', '201702'],
     ['--year', '2017'],
     ['--gps-start-time', '1170633618', '--gps-end-time', '1170720018']),
)
def test_main_loop_over_modes(tmpdir, caplog, mode):
    outdir = str(tmpdir)
    (global_, k1test, x509cert) = _get_inputs()
    args = [
        '--verbose',
        '--ifo', 'K1',
        '--universe', 'local',
        '--config-file', k1test,
        '--global-config', global_,
        '--single-process',
        '--output-dir', outdir,
    ]
    # test log output
    batch.main(args + mode)
    assert "Copied all INI configuration files to ./etc" in caplog.text
    assert " -- Configured HTML htmlnode job" in caplog.text
    assert " -- Configured job for config {}".format(
        os.path.join(outdir, "etc", os.path.basename(k1test))) in caplog.text
    assert "Setup complete, DAG written to: {}".format(
        os.path.join(outdir, "gw_summary_pipe.dag")) in caplog.text
    # clean up
    for filename in (global_, k1test, x509cert):
        os.remove(filename)
    shutil.rmtree(outdir, ignore_errors=True)


def test_main_invalid_modes(capsys):
    args = [
        '--ifo', 'V1',
        '--day', '20170209',
        '--month', '201702',
    ]
    # test output
    with pytest.raises(SystemExit):
        batch.main(args)
    (_, err) = capsys.readouterr()
    assert err.endswith("Please give only one of --day, --month, or "
                        "--gps-start-time and --gps-end-time.\n")



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Get range data
"""

import re

from gwpy import astro
from gwpy.frequencyseries import FrequencySeries
from gwpy.spectrogram import SpectrogramList
from gwpy.timeseries import TimeSeriesList

from .. import globalv
from ..utils import re_cchar
from ..channels import get_channel
from .utils import (use_segmentlist, make_globalv_key)
from .timeseries import (add_timeseries, get_timeseries)
from .spectral import (add_spectrogram, get_spectrogram)

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'


def _metadata(channel, **rangekwargs):
    """Return a common set of metadata for range calculations
    """
    channel = get_channel(channel)
    key = make_globalv_key(get_range_channel(channel, **rangekwargs))
    return (channel, key)


def _segments_diff(segments, havesegs, query=True):
    """Return a diff of two `SegmentList`, and determine whether to query for
    new data
    """
    new = segments - havesegs
    query &= abs(new) != 0
    return (new, query)


def get_range_channel(channel, **rangekwargs):
    """Return the meta-channel name used to store range data
    """
    channel = get_channel(channel)
    if rangekwargs:
        re_float = re.compile(r'[.-]')
        rkey = '_'.join(['%s_%s' % (re_cchar.sub('_', key),
                                    re_float.sub('_', str(val))) for key, val
                        in rangekwargs.items()])
        return '%s_%s' % (channel.ndsname, rkey)
    return channel.ndsname


@use_segmentlist
def get_range(channel, segments, config=None, cache=None,
              query=True, nds=None, return_=True, nproc=1,
              datafind_error='raise', frametype=None,
              stride=None, fftlength=None, overlap=None,
              method=None, **rangekwargs):
    """Calculate the sensitive distance for a given strain channel
    """
    channel, key = _metadata(channel, **rangekwargs)
    # get new segments
    havesegs = globalv.DATA.get(key, TimeSeriesList()).segments
    new, query = _segments_diff(segments, havesegs, query)
    if query:  # calculate new range
        spectrograms = get_spectrogram(
            channel, new, config=config, cache=cache, nds=nds, format='psd',
            frametype=frametype, nproc=nproc, datafind_error=datafind_error,
            stride=stride, fftlength=fftlength, overlap=overlap, method=method)
        for sg in spectrograms:  # calculate range for each spectrogram
            ts = astro.range_timeseries(sg, **rangekwargs)
            ts.channel = key
            add_timeseries(ts, key=key)

    if return_:
        return get_timeseries(key, segments, query=False)


@use_segmentlist
def get_range_spectrogram(channel, segments, config=None, cache=None,
                          query=True, nds=None, return_=True, nproc=1,
                          datafind_error='raise', frametype=None, stride=60,
                          fftlength=None, overlap=None, method=None,
                          **rangekwargs):
    """Estimate the spectral contribution to sensitive distance for a given
    strain channel
    """
    channel, key = _metadata(channel, **rangekwargs)
    # get new segments
    havesegs = globalv.SPECTROGRAMS.get(key, SpectrogramList()).segments
    new, query = _segments_diff(segments, havesegs, query)
    if query:  # calculate new data
        spectrograms = get_spectrogram(
            channel, new, config=config, cache=cache, nds=nds, format='psd',
            frametype=frametype, nproc=nproc, datafind_error=datafind_error,
            stride=stride, fftlength=fftlength, overlap=overlap, method=method)
        for sg in spectrograms:  # get contribution from each spectrogram
            outspec = astro.range_spectrogram(sg, **rangekwargs)
            outspec.channel = key
            add_spectrogram(outspec if 'energy' in rangekwargs else
                            outspec**(1/2.), key=key)

    if return_:
        return globalv.SPECTROGRAMS.get(key, SpectrogramList())


@use_segmentlist
def get_range_spectrum(channel, segments, config=None, cache=None, query=True,
                       nds=None, return_=True, nproc=1, datafind_error='raise',
                       frametype=None, stride=60, fftlength=None, overlap=None,
                       method=None, which='all', **rangekwargs):
    """Compute percentile spectra of the range integrand from a set of
    spectrograms
    """
    name = str(channel)
    cmin = '%s.min' % name
    cmax = '%s.max' % name

    if name not in globalv.SPECTRUM:
        speclist = get_range_spectrogram(
            channel, segments, config=config, cache=cache, query=query,
            nds=nds, return_=return_, nproc=nproc, frametype=frametype,
            datafind_error=datafind_error, method=method, stride=stride,
            fftlength=fftlength, overlap=overlap, **rangekwargs)
        specgram = speclist.join(gap='ignore')
        try:  # store median spectrum
            globalv.SPECTRUM[name] = specgram.percentile(50)
        except (ValueError, IndexError):
            unit = 'Mpc' if 'energy' in rangekwargs else 'Mpc^2 / Hz'
            globalv.SPECTRUM[name] = FrequencySeries(
                [], channel=channel, f0=0, df=1, unit=unit)
            globalv.SPECTRUM[cmin] = globalv.SPECTRUM[name]
            globalv.SPECTRUM[cmax] = globalv.SPECTRUM[name]
        else:  # store percentiles
            globalv.SPECTRUM[cmin] = specgram.percentile(5)
            globalv.SPECTRUM[cmax] = specgram.percentile(95)

    if not return_:
        return

    if which == 'all':
        return (globalv.SPECTRUM[name], globalv.SPECTRUM[cmin],
                globalv.SPECTRUM[cmax])
    if which == 'mean':
        return globalv.SPECTRUM[name]
    if which == 'min':
        return globalv.SPECTRUM[cmin]
    if which == 'max':
        return globalv.SPECTRUM[cmax]
    raise ValueError("Unrecognised value for `which`: %r" % which)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Test suite

"""

import pytest

from astropy import units

from gwpy.detector import ChannelList

from .. import (globalv, channels)
from ..mode import (get_mode, set_mode)

from gwpy.detector import Channel

from .common import empty_globalv_CHANNELS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

TEST_NAME = 'H1:GDS-CALIB_STRAIN'
TREND_NAME = 'H1:TEST-TREND_CHANNEL.rms,m-trend'
TREND_NAME2 = 'H1:TEST-TREND_CHANNEL.mean,m-trend'
TREND_NAME3 = 'H1:TEST-TREND_CHANNEL_3.mean'
TREND_NAME4 = 'H1:TEST-TREND_CHANNEL_4.mean'

DEFAULT_MODE = get_mode()


def teardown_module():
    """Undo any set_mode() operations from this module
    """
    set_mode(DEFAULT_MODE)


@empty_globalv_CHANNELS
def test_get_channel():
    """Test :func:`gwsumm.channels.get_channel`
    """
    nchan = len(globalv.CHANNELS)

    # test simple query
    chan = channels.get_channel(TEST_NAME)
    assert len(globalv.CHANNELS) == nchan + 1
    assert chan.name == TEST_NAME

    # make sure that querying again returns the same object
    chan2 = channels.get_channel(TEST_NAME)
    assert len(globalv.CHANNELS) == nchan + 1
    assert chan2 is chan


@empty_globalv_CHANNELS
def test_get_channel_trend():
    """Test get_channel for trends

    `get_channel` should query for the trend and the underlying
    raw channel
    """
    # test simple query
    nchan = len(globalv.CHANNELS)
    chan = channels.get_channel(TREND_NAME)
    assert len(globalv.CHANNELS) == nchan + 2

    # test that raw doesn't get built again
    chan = channels.get_channel(TREND_NAME2)
    assert len(globalv.CHANNELS) == nchan + 3

    # test that raw matches trend
    raw = channels.get_channel(TREND_NAME.split('.')[0])
    assert len(globalv.CHANNELS) == nchan + 3
    assert raw.name == TREND_NAME.split('.')[0]

    # test default trend type
    chan = channels.get_channel(TREND_NAME3)
    assert chan.type is None


@empty_globalv_CHANNELS
def test_get_channels():
    names = [TEST_NAME, TREND_NAME, TREND_NAME2]
    nchan = len(globalv.CHANNELS)
    chans = channels.get_channels(names)
    # trend channels auto create entry for the upstream channel so '+ 1'
    assert len(globalv.CHANNELS) == nchan + 3 + 1
    for name, chan in zip(names, chans):
        assert name == chan.ndsname


@empty_globalv_CHANNELS
def test_update_missing_channel_params():
    # define empty channel
    chan = channels.get_channel('X1:TEST:1')
    assert chan.unit is None

    # update using kwargs
    channels.update_missing_channel_params('X1:TEST:1', unit='meter')
    assert chan.unit == units.meter
    chan.unit = None

    # update from another channel
    c2 = Channel('X1:TEST:1', unit='V')
    channels.update_missing_channel_params(c2)
    assert chan.unit == units.volt


@pytest.mark.parametrize('cstr, clist', [
    ('X1:TEST,Y1:TEST,\nZ1:TEST', ['X1:TEST', 'Y1:TEST', 'Z1:TEST']),
    ('X1:TEST.rms,m-trend,Y1:TEST.mean,s-trend',
     ['X1:TEST.rms,m-trend', 'Y1:TEST.mean,s-trend']),
])
def test_split(cstr, clist):
    assert channels.split(cstr) == clist


@pytest.mark.parametrize('cstr, clist', [
    ('X1:TEST + Y1:TEST', ['X1:TEST', 'Y1:TEST']),
    ('X1:TEST.mean,m-trend * Y1:TEST + 1',
     ['X1:TEST.mean,m-trend', 'Y1:TEST']),
    ('G1:DER_DATA_H-rms - 4 + G1:DER_DATA_BLAH',
     ['G1:DER_DATA_H-rms', 'G1:DER_DATA_BLAH']),
])
def test_split_combination(cstr, clist):
    split = channels.split_combination(cstr)
    assert isinstance(split, ChannelList)
    assert list(map(str, split)), clist



# -*- coding: utf-8
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""HTML5-specific extensions
"""

import re
import os.path

from pathlib import Path
from urllib.parse import urlparse

from MarkupPy import markup
from markdown import markdown

# global variables

DOCTYPE = '<!DOCTYPE html>'

COMMENTS_JS = """
        (function() {
            var dsq = document.createElement('script');
            dsq.type = 'text/javascript'; dsq.async = true;
            dsq.src = '//' + disqus_shortname + '.disqus.com/embed.js';
            (document.getElementsByTagName('head')[0] ||
             document.getElementsByTagName('body')[0]).appendChild(dsq);
        })();
"""

OVERLAY_INSTRUCTIONS = """\
This tool is designed to help visually correlate multiple types of data by
laying transparent figures atop one another. To use it, follow these steps:

* From the main page(s), drag-and-drop figures onto the button with this icon:
  <i class="fas fa-layer-group"></i>
* Expand this dialog box and use the `Overlay` button (to the right) to render
  a new figure by overlaying the ones selected
* Add more figures using the steps above, click `Overlay` to re-render, and
  use the `Download` button to save to a local file
* Use the `Clear` button to clear all figure selections

**Note:** figures are remembered across tabs, so multiple times and subsystems
can be compared at once.\
"""


# -- utilities ----------------------------------------------------------------

def _expand_path(path):
    """Expand a server path that may contain symbolic links
    """
    subbed = Path(re.sub(r'^/\~(.*?)/', r'/home/\1/public_html/', path))
    resolved = subbed.resolve() if subbed.exists() else subbed
    return re.sub(r'^/home/(.*?)/public_html/', r'/~\1/',
                  str(resolved) if resolved.exists() else path)


def load_state(url):
    """Construct the HTML script required to load the Tab HTML
    for a given :class:`~gwsumm.state.core.SummaryState`

    Parameters
    ----------
    url : `str`
        path (relative to <base>) of HTML to load
    id_ : `str`, optional, default: '#main'
        <div> 'id' in which to load HTML

    Returns
    -------
    HTML : `str`
        HTML one-liner with script loading
    """
    id_ = os.path.splitext(os.path.basename(url))[0]
    page = markup.page()
    page.script()
    page.add('if (location.hash.length <= 1) {')
    page.add('    jQuery("#state_%s").load_state("%s");' % (id_, url))
    page.add('}')
    page.script.close()
    return page


def load(url, id_='main', error=False, success=None):
    """Construct the HTML script required to load a url into the
    HTML element with the given unique ``id_``.
    """
    ps = urlparse(url)
    if not ps.netloc and not error:
        return markup.given_oneliner.script('jQuery("#%s").load("%s");'
                                            % (id_, url))
    elif ps.netloc and not error:
        error = ('alert("Cannot load content from %r, use browser console '
                 'to inspect failure.");' % url)
    else:
        if not isinstance(error, (str, markup.page)):
            error = 'Failed to load content from %r' % url
        error = ('jQuery("#%s").html("<div class=\'alert alert-warning '
                 'text-justify shadow-sm\'><p>%s</p></div>");' % (id_, error))
    if success is None:
        success = 'jQuery("#%s").html(data);' % id_
    url = _expand_path(url)
    return markup.given_oneliner.script("""
    jQuery.ajax({
        url : '%s',
        type : 'GET',
        success: function(data, statusText, jqhxr){%s},
        error: function(xhr, status, error){%s}
        });\n""" % (url, success, error))


def comments_box(name, identifier=None, title=None, url=None):
    """Generate a Disqus comments box
    """
    page = markup.page()
    page.div(id_='disqus_thread')
    page.script(type='text/javascript')
    page.add('    var disqus_shortname = "%s";' % name)
    if identifier:
        page.add('    var disqus_identifier = "%s";' % identifier)
    if title:
        page.add('    var disqus_title = "%s";' % title)
    if url:
        page.add('    var disqus_url = "%s";' % url)
    page.add(COMMENTS_JS)
    page.script.close()
    page.noscript("Please enable JavaScript to view the")
    page.a("comments powered by Disqus",
           href="https://disqus.com/?ref_noscript")
    page.div.close()
    return page


def ldvw_qscan(channel, time, fmin=10, fmax='inf', qmin=4, qmax=100):
    """Generate a Q-scan through LIGO DataViewer Web (LDVW)
    """
    channel = str(channel)
    if isinstance(time, (tuple, list)):
        label = 'Launch omega scans'
        title = 'Batch-process omega scans of the loudest triggers via LDVW'
        times = '&'.join('wdq_gps=' + str(t) for t in time)
        query = ('Wdq?submitAct=go&wdq_ifo=%s&wdq_cmap=viridis&%s&'
                 'wdq_prog=py-Omega&goBtn=goBtn') % (channel[:2], times)
    else:
        label = 'Q-scan'
        title = 'Q-scan {0} at {1} via LDVW'.format(channel, time)
        query = ('view?act=doplot&chanName={0}&doQxfrm=yes&strtTime={1}&'
                 '&qxfrm_pltfrq={2} {3}&qxfrm_srchqrng={4} {5}&'
                 'qxfrm_plttimes=0.5 2 8').format(
                     channel, time, fmin, fmax, qmin, qmax)
    uri = 'https://ldvw.ligo.caltech.edu/ldvw/{0}'.format(query)
    return markup.oneliner.a(
        label, href=uri, target='_blank', rel='external',
        class_='btn btn-outline-secondary btn-sm', title=title)


def dialog_box(content, title, id_, btntxt):
    """Generate a dialog box to be loaded modal atop the main page

    Parameters
    ----------
    content : `str`
        either raw markdown text or the path to a file containing markdown,
        this will be rendered in HTML as the contents of the dialog box

    title : `str`
        title to display atop the dialog box

    id_ : `str`
        unique identifier for the dialog box

    btntxt : `str`
        text (usually a single character) to appear inside a sticky button
        that opens the dialog box

    Returns
    -------
    page : `~MarkupPy.markup.page`
        fully rendered HTML containing the dialog box
    """
    btnargs = {
        'title': title,
        'id_': '-'.join([id_, 'btn']),
        'class_': 'btn-float btn-open shadow',
        'data-id': '#' + id_,
    }
    page = markup.page()
    page.button(btntxt, **btnargs)
    page.div(
        class_='dialog',
        title=title,
        id_=id_,
    )
    if os.path.isfile(content):
        with open(content, 'r') as source:
            content = source.read()
    page.add(markdown(str(content)))
    page.div.close()
    return page


def overlay_canvas():
    """Generate a dialog box allowing users to select and overlay plots

    Returns
    -------
    page : `~MarkupPy.markup.page`
        fully rendered HTML containing the dialog box
    """
    page = markup.page()
    page.h1('Overlay figures for easy comparison')
    page.hr(class_='row-divider')
    page.div(class_='row', id_='overlay-outer')
    page.div(class_='col-md-4')
    page.div(class_='card card-body shadow-sm', id_='overlay-info')
    page.h4('Instructions')
    page.add(markdown(OVERLAY_INSTRUCTIONS))
    page.div.close()  # card card-body shadow-sm
    page.div.close()  # col-md-4
    page.div(class_='col-md-8')
    page.div(class_='text-center')
    page.a('Overlay', title='Overlay all selected figures',
           class_='btn btn-light shadow-sm', id_='overlay-figures')
    page.a('Download', title='Download overlay figure',
           class_='btn btn-light shadow-sm', id_='download-overlay')
    page.a('Clear', title='Clear all figure selections',
           class_='btn btn-light shadow-sm', id_='clear-figures')
    page.div.close()  # text-center
    page.br()
    page.add(markup.oneliner.canvas(id_='overlay-canvas'))
    page.div.close()  # col-md-8
    page.div.close()  # row
    return dialog_box(
        str(page), title='Overlay figures', id_='overlay',
        btntxt=markup.oneliner.i('', class_='fas fa-layer-group'))



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2020)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwsumm.plot.triggers` command-line interface
"""

import os
import pytest
import shutil

from unittest import mock

from gwpy.segments import (
    Segment,
    SegmentList,
    DataQualityFlag,
)

from .... import globalv
from .. import __main__ as triggers_cli

__author__ = 'Alex Urban <alexander.urban@ligo.org>'

# -- test configuration

CHANNEL = "H1:GDS-CALIB_STRAIN"

# -- test data

LOCK = DataQualityFlag(
    name="H1:DMT-GRD_ISC_LOCK_NOMINAL:1",
    active=SegmentList([Segment(2, 2048)]),
    known=SegmentList([Segment(0, 3600)]),
)


# -- cli tests ----------------------------------------------------------------

@mock.patch(
    'gwpy.segments.DataQualityFlag.query_dqsegdb',
    return_value=LOCK,
)
def test_main(dqflag, tmpdir, caplog):
    outdir = str(tmpdir)
    plot = os.path.join(outdir, "triggers.png")
    args = [
        CHANNEL,
        '0', '3600',
        '--snr', '1',
        '--state', LOCK.name,
        '--output-file', plot,
    ]
    # test output
    with pytest.warns(UserWarning) as record:
        triggers_cli.main(args)
    assert os.path.exists(plot)
    assert len(os.listdir(outdir)) == 1
    assert 'Read 0 events' in caplog.text
    assert "0 events in state '{}'".format(LOCK.name) in caplog.text
    assert '0 events remaining with snr >= 1.0' in caplog.text
    assert 'Plot saved to {}'.format(plot) in caplog.text
    # test the `UserWarning`
    # FIXME: once MatplotlibDeprecationWarning about colormaps is fixed,
    #        assert that this UserWarning is the **only** warning
    assert (record[0].message.args[0] ==
            "Caught ValueError: No channel-level directory found at "
            "/home/detchar/triggers/*/H1/GDS-CALIB_STRAIN_Omicron. Either "
            "the channel name or ETG names are wrong, or this channel is not "
            "configured for this ETG.")
    # clean up
    globalv.TRIGGERS = {}
    shutil.rmtree(outdir, ignore_errors=True)


def test_main_with_cache_and_tiles(tmpdir, caplog):
    outdir = str(tmpdir)
    cache = os.path.join(outdir, "empty.cache")
    plot = os.path.join(outdir, "triggers.png")
    args = [
        CHANNEL,
        '0', '3600',
        '--cache-file', cache,
        '--snr', '1',
        '--plot-params', 'legend-loc="upper right"',
        '--tiles',
        '--output-file', plot,
    ]
    # write an empty cache file
    with open(cache, 'w') as f:
        f.write("")
    # test output
    triggers_cli.main(args)
    assert os.path.exists(plot)
    assert len(os.listdir(outdir)) == 2  # 1 input, 1 output
    assert 'Read cache of 0 files' in caplog.text
    assert 'Read 0 events' in caplog.text
    assert '0 events remaining with snr >= 1.0' in caplog.text
    assert 'Plot saved to {}'.format(plot) in caplog.text
    # clean up
    globalv.TRIGGERS = {}
    shutil.rmtree(outdir, ignore_errors=True)


def test_main_invalid_columns(capsys):
    args = [
        CHANNEL,
        '0', '3600',
        '--columns', 'invalid',
    ]
    # test output
    with pytest.raises(SystemExit):
        triggers_cli.main(args)
    (_, err) = capsys.readouterr()
    assert err.endswith("--columns must receive at least two columns, got 1\n")



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for channel access
"""

import re
from functools import wraps

from astropy.units import Unit

from gwpy.detector import (Channel, ChannelList)
from gwpy.io.nds2 import Nds2ChannelType

from . import globalv
from .utils import re_quote

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

NDS2_TYPES = Nds2ChannelType.names()
CIS_URL = 'https://cis.ligo.org'
re_channel = re.compile(r'[A-Z]\d:[a-zA-Z0-9]+'  # core channel section L1:TEST
                        r'(?:[-_][a-zA-Z0-9_]+)?'  # underscore-delimited parts
                        r'(?:\.[a-z]+)?'  # trend type
                        r'(?:,[a-z-]+)?')  # NDS channel type


# -- channel creation ---------------------------------------------------------

def _match(channel):
    channel = Channel(channel)
    name = str(channel)
    type_ = channel.type
    found = globalv.CHANNELS.sieve(
        name=name, type=type_, exact_match=True)

    # if match, return now
    if found:
        return found

    # if no matches, try again without matching type
    found = globalv.CHANNELS.sieve(name=name, exact_match=True)
    if len(found) == 1:
        # found single match that is less specific, so we make it more
        # specific. If someone else wants another type for the sme channel
        # this is fine, and it will create another channel.
        found[0].type = type_
    return found


def _find_parent(channel):
    """Find the parent for a given channel

    This is either the raw channel from which a trend was generated, or the
    real channel for a mathematically-manipulated channel.

    Raises
    ------
    ValueError
        if the parent cannot be parsed
    """
    if channel.trend:
        parent = str(channel).rsplit('.')[0]
    else:
        parent, = re_channel.findall(str(channel))
    if parent == str(channel):
        raise ValueError("Cannot find parent for '{0!s}'".format(channel))
    return get_channel(parent)


def _update_dependent(channel):
    """Update a trend channel from its parent
    """
    try:
        source = _find_parent(channel)
    except (ValueError, IndexError):
        return channel
    if source is channel:
        return channel

    channel.url = source.url
    channel.unit = source.unit

    # copy custom named params
    #     this works because the upstream Channel stores all
    #     attributes in private variable names
    # NOTE: we need to exclude 'resample' to not attempt to upsample trends
    for param in filter(
            lambda x: not x.startswith('_') and
            not hasattr(channel, x) and x not in ('resample',),
            vars(source)):
        try:
            setattr(channel, param, getattr(source, param))
        except AttributeError:
            pass
    return channel


def _with_update_dependent(func):
    """Decorate ``func`` to call `_update_dependent()` upon exit
    """
    @wraps(func)
    def wrapped_func(*args, **kwargs):
        _update = kwargs.pop('find_parent', True)
        out = func(*args, **kwargs)
        if _update and out.trend:
            out = _update_dependent(out)
        return out
    return wrapped_func


def _new(channel, find_parent=True):
    """Create a new `~gwpy.detector.Channel` in the globalv cache.
    """
    # convert to Channel
    if isinstance(channel, Channel):
        new = channel
    else:
        new = Channel(channel)
    name = str(channel)
    type_ = new.type

    # work out what kind of channel it is
    parts = re_channel.findall(name)

    # match single raw channel for LIGO
    if (
            len(parts) == 1 and
            new.ifo in ('H1', 'L1') and
            not re.search(r'\.[a-z]+\Z', name)
    ):
        new.url = '%s/channel/byname/%s' % (CIS_URL, str(new))

    # match single trend
    elif len(parts) == 1:
        # set default trend type based on mode
        if type_ is None and ':DMT-' in name:  # DMT is always m-trend
            new.type = 'm-trend'
        # match parameters from 'raw' version of this channel

    # match composite channel
    else:
        new.subchannels = parts
        new._ifo = "".join(set(p.ifo for p in map(Channel, parts) if p.ifo))

    if find_parent:
        _update_dependent(new)

    # store new channel and return
    globalv.CHANNELS.append(new)
    try:
        return get_channel(new)
    except RuntimeError as e:
        if 'maximum recursion depth' in str(e):
            raise RuntimeError("Recursion error while accessing channel "
                               "information for %s" % str(channel))
        raise


@_with_update_dependent
def get_channel(channel, find_parent=True, timeout=5):
    """Find (or create) a :class:`~gwpy.detector.Channel`.

    If ``channel`` has already been created, the cached copy will be
    returned, otherwise a new `~gwpy.detector.Channel` object will be created.

    Parameters
    ----------
    channel : `str`
        name of new channel

    find_parent : `bool`, optional, default: `True`
        query for raw version of trend channel (trends not in CIS)

    timeout : `float`, optional, default: `5`
        number of seconds to wait before connection times out

    Returns
    -------
    Channel : :class:`~gwpy.detector.Channel`
        new channel.
    """
    chans = re_channel.findall(str(channel))
    nchans = len(chans)

    # match compound channel name
    if nchans > 1 or (nchans == 1 and chans[0] != str(channel)):
        name = str(channel)
        # handle special characters in channel name
        rename = name
        for cchar in ['+', '*', '^', '|']:
            rename = rename.replace(cchar, r'\%s' % cchar)
        found = globalv.CHANNELS.sieve(name=rename, exact_match=True)
    # match normal channel
    else:
        found = _match(channel)

    # if single match, return it
    if len(found) == 1:
        return found[0]

    # if multiple matches raise error (unless weird circumstances)
    elif len(found) > 1:
        cstrings = set(['%s [%s, %s]' % (c.ndsname, c.sample_rate, c.unit)
                        for c in found])
        if len(cstrings) == 1:  # somehow all channels are the same
            return found[0]
        raise ValueError("Ambiguous channel request '%s', multiple "
                         "existing channels recovered:\n    %s"
                         % (str(channel), '\n    '.join(cstrings)))

    # otherwise there were no matches, so we create a new channel
    return _new(channel, find_parent=find_parent)


def get_channels(channels, **kwargs):
    """Find (or create) multiple channels.

    See Also
    --------
    get_channel
    """
    return ChannelList(get_channel(c, **kwargs) for c in channels)


# -- channel manipulation -----------------------------------------------------

def update_missing_channel_params(channel, **kwargs):
    """Update empty channel parameters using the given input

    This method will only set parameters in the channel if the target
    parameter is `None`.

    Parameters
    ----------
    channel : `~gwpy.detector.Channel`
        channel to update
    **kwargs
        `(key, value)` pairs to set
    """
    target = get_channel(str(channel))
    if isinstance(channel, Channel):
        for param in ['unit', 'sample_rate', 'frametype']:
            if getattr(target, param) is None or (
                    param == 'unit' and
                    getattr(target, param) is Unit('undef')):
                setattr(target, param, getattr(channel, param))
    for param in kwargs:
        if getattr(target, param) is None:
            setattr(target, param, kwargs[param])
    return target


def update_channel_params():
    """Update the `globalv.CHANNELS` list based on internal parameter changes

    This is required to update `Channel.type` based on `Channel.frametype`,
    and similar.
    """
    for c in globalv.CHANNELS:
        # update type based on frametype
        if c.type is None and c.frametype == '{0.ifo}_M'.format(c):
            c.type = 'm-trend'
        elif c.type is None and c.frametype == '{0.ifo}_T'.format(c):
            c.type = 's-trend'

        # update sample_rate based on trend type
        if c.type == 'm-trend' and c.sample_rate is None:
            c.sample_rate = 1/60.
        elif c.type == 's-trend' and c.sample_rate is None:
            c.sample_rate = 1.
    return


# -- string parsing -----------------------------------------------------------

def split(channelstring):
    """Split a comma-separated list of channels that may, or may not
    contain NDS2 channel types as well
    """
    out = []
    channelstring = re_quote.sub('', channelstring)
    while True:
        channelstring = channelstring.strip('\' \n')
        if ',' not in channelstring:
            break
        # check for complete line without NDS type
        line = channelstring.split('\n')[0].rstrip('\', \n')
        if ',' not in line:
            try:
                channelstring = channelstring.split('\n', 1)[1]
            except IndexError:
                pass
            else:
                out.append(line)
                continue
        # check for channel name with optional nds type
        for nds2type in NDS2_TYPES + ['']:
            if nds2type and ',%s' % nds2type in channelstring:
                try:
                    channel, ctype, channelstring = channelstring.split(',', 2)
                except ValueError:
                    channel, ctype = channelstring.split(',')
                    channelstring = ''
                out.append('%s,%s' % (channel, ctype))
                break
            elif nds2type == '' and ',' in channelstring:
                channel, channelstring = channelstring.split(',', 1)
                out.append(channel)
                break
    if channelstring:
        out.append(channelstring)
    return out


def split_combination(channelstring):
    """Split a math-combination of channels

    Returns
    -------
    channels : `~gwpy.detector.ChannelList`
    """
    return get_channels(re_channel.findall(str(channelstring)),
                        find_parent=False)



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwsumm.data`

"""

import os.path
import operator
import tempfile
import shutil
from collections import OrderedDict
from urllib.request import urlopen

import pytest

from numpy import (arange, testing as nptest)

from lal.utils import CacheEntry

from glue.lal import Cache

from gwpy.timeseries import TimeSeries
from gwpy.detector import Channel
from gwpy.segments import (Segment, SegmentList)

from .. import (data, globalv)
from ..data import (utils, mathutils)

from .common import empty_globalv_CHANNELS

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'

LOSC_DATA = {
    'H1:LOSC-STRAIN': ['https://losc.ligo.org/s/events/GW150914/'
                       'H-H1_LOSC_4_V1-1126259446-32.gwf'],
    'L1:LOSC-STRAIN': ['https://losc.ligo.org/s/events/GW150914/'
                       'L-L1_LOSC_4_V1-1126259446-32.gwf'],
}
LOSC_SEGMENTS = SegmentList([Segment(1126259446, 1126259478)])


def download(remote, target=None):
    """Download a file
    """
    if target is None:
        suffix = os.path.splitext(remote)[1]
        _, target = tempfile.mkstemp(suffix=suffix, prefix='gwsumm-tests-')
    response = urlopen(remote)
    with open(target, 'wb') as f:
        f.write(response.read())
    return target


class TestData(object):
    """Tests for :mod:`gwsumm.data`:
    """
    @classmethod
    def setup_class(cls):
        cls.FRAMES = {}
        cls._tempdir = tempfile.mkdtemp(prefix='gwsumm-test-data-')
        # get data
        for channel in LOSC_DATA:
            cls.FRAMES[channel] = Cache()
            for gwf in LOSC_DATA[channel]:
                target = os.path.join(cls._tempdir, os.path.basename(gwf))
                download(gwf, target)
                cls.FRAMES[channel].append(CacheEntry.from_T050017(target))

    @classmethod
    def teardown_class(cls):
        # remove the temporary data
        shutil.rmtree(cls._tempdir)

    # -- test utilities -------------------------

    def test_find_frame_type(self):
        channel = Channel('L1:TEST-CHANNEL')
        assert data.find_frame_type(channel) == 'L1_R'

        channel = Channel('C1:TEST-CHANNEL')
        assert data.find_frame_type(channel) == 'R'

        channel = Channel('H1:TEST-CHANNEL.rms,s-trend')
        assert data.find_frame_type(channel) == 'H1_T'

        channel = Channel('H1:TEST-CHANNEL.rms,m-trend')
        assert data.find_frame_type(channel) == 'H1_M'

        channel = Channel('H1:TEST-CHANNEL.rms,online')
        assert data.find_frame_type(channel) == 'H1_lldetchar'

    def test_get_channel_type(self):
        assert data.get_channel_type('L1:TEST-CHANNEL') == 'adc'
        assert data.get_channel_type('G1:DER_DATA_HL') == 'proc'
        assert data.get_channel_type('H1:GDS-CALIB_STRAIN') == 'proc'
        assert data.get_channel_type('V1:GDS-CALIB_STRAIN') == 'adc'

    @empty_globalv_CHANNELS
    def test_make_globalv_key(self):
        fftparams = utils.get_fftparams(
            'L1:TEST-CHANNEL',
            stride=123.456,
            window='test-window',
            method='scipy-welch',
        )
        key = utils.make_globalv_key('L1:TEST-CHANNEL', fftparams)
        assert key == ';'.join([
            'L1:TEST-CHANNEL',  # channel
            'scipy-welch',  # method
            '',  # fftlength
            '',  # overlap
            'test-window',  # window
            '123.456',  # stride
            '',  # FFT scheme
        ])

    def test_get_fftparams(self):
        fftparams = utils.get_fftparams('L1:TEST-CHANNEL')
        assert isinstance(fftparams, utils.FftParams)

        for key in utils.FFT_PARAMS:
            assert (getattr(fftparams, key) is
                    utils.DEFAULT_FFT_PARAMS.get(key, None))

        fftparams = utils.get_fftparams('L1:TEST-CHANNEL', window='hanning',
                                        overlap=0)
        assert fftparams.window == 'hanning'
        assert fftparams.overlap == 0

        with pytest.raises(ZeroDivisionError):
            utils.get_fftparams(None, stride=0)

    @pytest.mark.parametrize('definition, math', [
        (
             'L1:TEST + L1:TEST2',
             ([('L1:TEST', 'L1:TEST2'), ([], [])], [operator.add]),
        ),
        (
             'L1:TEST + L1:TEST2 * 2',
             ([('L1:TEST', 'L1:TEST2'), ([], [(operator.mul, 2)])],
              [operator.add]),
        ),
        (
             'L1:TEST * 2 + L1:TEST2 ^ 5',
             ([('L1:TEST', 'L1:TEST2'),
               ([(operator.mul, 2)], [(operator.pow, 5)])],
              [operator.add]),
        ),
    ])
    def test_parse_math_definition(self, definition, math):
        chans, operators = mathutils.parse_math_definition(definition)
        assert chans == OrderedDict(list(zip(*math[0])))
        assert operators == math[1]

    # -- test add/get methods -------------------

    def test_add_timeseries(self):
        a = TimeSeries([1, 2, 3, 4, 5], name='test name', epoch=0,
                       sample_rate=1)

        # test simple add using 'name'
        data.add_timeseries(a)
        assert 'test name' in globalv.DATA
        assert len(globalv.DATA['test name']) == 1
        assert globalv.DATA['test name'][0] is a

        # test add using key kwarg
        data.add_timeseries(a, key='test key')
        assert globalv.DATA['test key'][0] is a

        # test add to existing key with coalesce
        b = TimeSeries([6, 7, 8, 9, 10], name='test name 2', epoch=5,
                       sample_rate=1)
        data.add_timeseries(b, key='test key', coalesce=True)
        assert len(globalv.DATA['test key']) == 1
        nptest.assert_array_equal(globalv.DATA['test key'][0].value,
                                  arange(1, 11))

    def test_get_timeseries(self):
        # empty globalv.DATA
        globalv.DATA = type(globalv.DATA)()

        # test simple get after add
        a = TimeSeries([1, 2, 3, 4, 5], name='test name', epoch=0,
                       sample_rate=1)
        data.add_timeseries(a)
        b, = data.get_timeseries('test name', [(0, 5)], nproc=1)
        nptest.assert_array_equal(a.value, b.value)
        assert a.sample_rate.value == b.sample_rate.value

        # test more complicated add with a cache
        a, = data.get_timeseries('H1:LOSC-STRAIN', LOSC_SEGMENTS,
                                 cache=self.FRAMES['H1:LOSC-STRAIN'],
                                 nproc=1)
        b, = data.get_timeseries('H1:LOSC-STRAIN', LOSC_SEGMENTS,
                                 nproc=1)
        nptest.assert_array_equal(a.value, b.value)

    @empty_globalv_CHANNELS
    def test_get_spectrogram(self):
        with pytest.raises(TypeError):
            data.get_spectrogram('H1:LOSC-STRAIN', LOSC_SEGMENTS,
                                 cache=self.FRAMES['H1:LOSC-STRAIN'],
                                 nproc=1)
        data.get_spectrogram('H1:LOSC-STRAIN', LOSC_SEGMENTS,
                             cache=self.FRAMES['H1:LOSC-STRAIN'],
                             stride=4, fftlength=2, overlap=1,
                             nproc=1)

    def test_get_spectrum(self):
        a, _, _ = data.get_spectrum('H1:LOSC-STRAIN', LOSC_SEGMENTS,
                                    cache=self.FRAMES['H1:LOSC-STRAIN'],
                                    nproc=1)
        b, _, _ = data.get_spectrum('H1:LOSC-STRAIN', LOSC_SEGMENTS,
                                    format='asd',
                                    cache=self.FRAMES['H1:LOSC-STRAIN'],
                                    nproc=1)
        nptest.assert_array_equal(a.value ** (1/2.), b.value)

    def test_get_coherence_spectrogram(self):
        cache = Cache([e for c in self.FRAMES for e in self.FRAMES[c]])
        data.get_coherence_spectrogram(
            ('H1:LOSC-STRAIN', 'L1:LOSC-STRAIN'), LOSC_SEGMENTS, cache=cache,
            stride=4, fftlength=2, overlap=1, nproc=1,
        )

    def test_get_coherence_spectrum(self):
        cache = Cache([e for c in self.FRAMES for e in self.FRAMES[c]])
        data.get_coherence_spectrogram(
            ('H1:LOSC-STRAIN', 'L1:LOSC-STRAIN'), LOSC_SEGMENTS, cache=cache,
            stride=4, fftlength=2, overlap=1, nproc=1,
        )



# -*- coding: utf-8 -*-
# Copyright (C) Duncan Macleod (2013)
#
# This file is part of GWSumm.
#
# GWSumm is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWSumm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWSumm.  If not, see <http://www.gnu.org/licenses/>.

"""A `SummaryState` defines a sub-set of time over which a `~gwsumm.tabs.Tab`
should be processed.
Each `SummaryState` is normally tied to one or more data-quality flags marking
times during which each of the LIGO instruments was operating in a certain
configuration, or was subject to a known noise interference.

==================
The state registry
==================

GWSumm defines a state 'registry', simply a record of all `SummaryState`
objects that have been defined (and registered) so far in a given program.
The registry just makes remembering states in complicated programs a little
easier.

Any `SummaryState` can be registered with an arbitrary name as follows::

    >>> from gwsumm.state.registry import register_state
    >>> register_state(mystate, 'my state')

and can be recovered later::

    >>> from gwsumm.state.registry import get_state
    >>> mystate = get_state('my state')

=============
API reference
=============

.. autosummary::
   :toctree: api

   SummaryState
   get_state
   get_states
   register_state

"""

from .core import SummaryState
from .registry import (get_state, get_states, register_state)
from .all import (ALLSTATE, generate_all_state)

__all__ = ['ALLSTATE', 'SummaryState', 'get_state', 'get_states',
           'register_state', 'generate_all_state']
